__author__ = 'saeedamen'

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import numpy

from finmarketpy.util.marketconstants import MarketConstants

from findatapy.util import SwimPool
from findatapy.util import LoggerManager

import pickle
#import zlib
#import lz4framed    # conda install -c conda-forge py-lz4framed
import blosc
import pickle

market_constants = MarketConstants()


class Backtest(object):
    """Conducts backtest for strategies trading assets. Assumes we have an input of total returns. Reports historical return statistics
    and returns time series.

    """

    def __init__(self):
        self._pnl = None
        self._portfolio = None
        return

    def calculate_diagnostic_trading_PnL(self, asset_a_df, signal_df, further_df=[], further_df_labels=[]):
        """Calculates P&L table which can be used for debugging purposes,

        The table is populated with asset, signal and further dataframes provided by the user, can be used to check signalling methodology.
        It does not apply parameters such as transaction costs, vol adjusment and so on.

        Parameters
        ----------
        asset_a_df : DataFrame
            Asset prices

        signal_df : DataFrame
            Trade signals (typically +1, -1, 0 etc)

        further_df : DataFrame
            Further dataframes user wishes to output in the diagnostic output (typically inputs for the signals)

        further_df_labels
            Labels to append to the further dataframes

        Returns
        -------
        DataFrame with asset, trading signals and returns of the trading strategy for diagnostic purposes

        """
        calculations = Calculations()
        asset_rets_df = calculations.calculate_returns(asset_a_df)
        strategy_rets = calculations.calculate_signal_returns(signal_df, asset_rets_df)

        reset_points = ((signal_df - signal_df.shift(1)).abs())

        asset_a_df_entry = asset_a_df.copy(deep=True)
        asset_a_df_entry[reset_points == 0] = numpy.nan
        asset_a_df_entry = asset_a_df_entry.ffill()

        asset_a_df_entry.columns = [x + '_entry' for x in asset_a_df_entry.columns]
        asset_rets_df.columns = [x + '_asset_rets' for x in asset_rets_df.columns]
        strategy_rets.columns = [x + '_strat_rets' for x in strategy_rets.columns]
        signal_df.columns = [x + '_final_signal' for x in signal_df.columns]

        for i in range(0, len(further_df)):
            further_df[i].columns = [x + '_' + further_df_labels[i] for x in further_df[i].columns]

        flatten_df = [asset_a_df, asset_a_df_entry, asset_rets_df, strategy_rets, signal_df]

        for f in further_df:
            flatten_df.append(f)

        return calculations.pandas_outer_join(flatten_df)

    def calculate_trading_PnL(self, br, asset_a_df, signal_df, contract_value_df, run_in_parallel):
        """Calculates P&L of a trading strategy and statistics to be retrieved later

        Calculates the P&L for each asset/signal combination and also for the finally strategy applying appropriate
        weighting in the portfolio, depending on predefined parameters, for example:
            static weighting for each asset
            static weighting for each asset + vol weighting for each asset
            static weighting for each asset + vol weighting for each asset + vol weighting for the portfolio

        Parameters
        ----------
        br : BacktestRequest
            Parameters for the backtest specifying start date, finish data, transaction costs etc.

        asset_a_df : pandas.DataFrame
            Asset prices to be traded

        signal_df : pandas.DataFrame
            Signals for the trading strategy

        contract_value_df : pandas.DataFrame
            Daily size of contracts
        """

        calculations = Calculations()
        risk_engine = RiskEngine()

        # # do an outer join first, so can fill out signal and fill it down
        # # this captures the case where the signal changes on an asset holiday
        # # it will just get delayed till the next tradable day when we do this
        # asset_df_2, signal_df_2 = asset_a_df.align(signal_df, join='outer', axis='index')
        # signal_df = signal_df_2.fillna(method='ffill')
        #
        # # now make sure the dates of both traded asset and signal are aligned properly
        # # and use as reference only those days where we have asset information
        # asset_df, signal_df = asset_a_df.align(signal_df, join='left', axis = 'index')

        logger = LoggerManager().getLogger(__name__)

        logger.info("Calculating trading P&L...")

        signal_df = signal_df.shift(br.signal_delay)
        asset_df, signal_df = calculations.join_left_fill_right(asset_a_df, signal_df)

        if (contract_value_df is not None):
            asset_df, contract_value_df = asset_df.align(contract_value_df, join='left', axis='index')
            contract_value_df = contract_value_df.fillna(
                method='ffill')  # fill down asset holidays (we won't trade on these days)

        # non-trading days of the assets (this may of course vary between the assets we are trading
        # if they are from different asset classes)
        non_trading_days = numpy.isnan(asset_df.values)

        # only allow signals to change on the days when we can trade assets
        signal_df = signal_df.mask(non_trading_days)  # fill asset holidays with NaN signals
        signal_df = signal_df.fillna(method='ffill')  # fill these down

        tc = br.spot_tc_bp

        signal_cols = signal_df.columns.values
        asset_df_cols = asset_df.columns.values

        pnl_cols = []

        for i in range(0, len(asset_df_cols)):
            pnl_cols.append(asset_df_cols[i] + " / " + signal_cols[i])

        asset_df = asset_df.fillna(method='ffill')  # fill down asset holidays (we won't trade on these days)
        returns_df = calculations.calculate_returns(asset_df)

        # apply a stop loss/take profit to every trade if this has been specified
        # do this before we start to do vol weighting etc.
        if br.take_profit is not None and br.stop_loss is not None:
            returns_df = calculations.calculate_returns(asset_df)

            # makes assumption that signal column order matches that of returns
            temp_strategy_rets_df = calculations.calculate_signal_returns_as_matrix(signal_df, returns_df)

            trade_rets_df = calculations.calculate_cum_rets_trades(signal_df, temp_strategy_rets_df)

            # pre_signal_df = signal_df.copy()

            signal_df = calculations.calculate_risk_stop_signals(signal_df, trade_rets_df, br.stop_loss, br.take_profit)

            # make sure we can't trade where asset price is undefined and carry over signal
            signal_df = signal_df.mask(non_trading_days)  # fill asset holidays with NaN signals
            signal_df = signal_df.fillna(method='ffill')  # fill these down (when asset is not trading

            # for debugging purposes
            # if True:
            #     signal_df_copy = signal_df.copy()
            #     trade_rets_df_copy = trade_rets_df.copy()
            #
            #     asset_df_copy.columns = [x + '_asset' for x in temp_strategy_rets_df.columns]
            #     temp_strategy_rets_df.columns = [x + '_strategy_rets' for x in temp_strategy_rets_df.columns]
            #     signal_df_copy.columns = [x + '_final_signal' for x in signal_df_copy.columns]
            #     trade_rets_df_copy.columns = [x + '_cum_trade' for x in trade_rets_df_copy.columns]
            #
            #     to_plot = calculations.pandas_outer_join([asset_df_copy, pre_signal_df, signal_df_copy, trade_rets_df_copy, temp_strategy_rets_df])
            #     to_plot.to_csv('test.csv')

        # do we have a vol target for individual signals?
        if br.signal_vol_adjust is True:
            leverage_df = risk_engine.calculate_leverage_factor(returns_df, br.signal_vol_target,
                                                                br.signal_vol_max_leverage,
                                                                br.signal_vol_periods, br.signal_vol_obs_in_year,
                                                                br.signal_vol_rebalance_freq,
                                                                br.signal_vol_resample_freq,
                                                                br.signal_vol_resample_type,
                                                                period_shift=br.signal_vol_period_shift)

            signal_df = pandas.DataFrame(
                signal_df.values * leverage_df.values, index=signal_df.index, columns=signal_df.columns)

            self._individual_leverage = leverage_df  # contains leverage of individual signal (before portfolio vol target)

        _pnl = calculations.calculate_signal_returns_with_tc_matrix(signal_df, returns_df, tc=tc)
        _pnl.columns = pnl_cols

        adjusted_weights_matrix = None

        # portfolio is average of the underlying signals: should we sum them or average them or use another
        # weighting scheme?
        if br.portfolio_combination is not None:
            if br.portfolio_combination == 'sum' and br.portfolio_combination_weights is None:
                portfolio = pandas.DataFrame(data=_pnl.sum(axis=1), index=_pnl.index, columns=['Portfolio'])
            elif br.portfolio_combination == 'mean' and br.portfolio_combination_weights is None:
                portfolio = pandas.DataFrame(data=_pnl.mean(axis=1), index=_pnl.index, columns=['Portfolio'])

                adjusted_weights_matrix = self.create_portfolio_weights(br, _pnl, method='mean')
            elif 'weighted' in br.portfolio_combination and isinstance(br.portfolio_combination_weights, dict):

                # get the weights for each asset
                adjusted_weights_matrix = self.create_portfolio_weights(br, _pnl, method=br.portfolio_combination)

                portfolio = pandas.DataFrame(data=(_pnl.values * adjusted_weights_matrix), index=_pnl.index)
                is_all_na = pandas.isnull(portfolio).all(axis=1)
                portfolio = pandas.DataFrame(portfolio.sum(axis=1), columns=['Portfolio'])

                # overwrite days when every asset PnL was null is NaN with nan
                portfolio[is_all_na] = numpy.nan
        else:
            portfolio = pandas.DataFrame(data=_pnl.mean(axis=1), index=_pnl.index, columns=['Portfolio'])

            adjusted_weights_matrix = self.create_portfolio_weights(br, _pnl, method='mean')

        portfolio_leverage_df = pandas.DataFrame(data=numpy.ones(len(_pnl.index)), index=_pnl.index,
                                                 columns=['Portfolio'])

        # should we apply vol target on a portfolio level basis?
        if br.portfolio_vol_adjust is True:
            # calculate portfolio leverage
            portfolio_leverage_df = risk_engine.calculate_leverage_factor(portfolio,
                                                                          br.portfolio_vol_target,
                                                                          br.portfolio_vol_max_leverage,
                                                                          br.portfolio_vol_periods,
                                                                          br.portfolio_vol_obs_in_year,
                                                                          br.portfolio_vol_rebalance_freq,
                                                                          br.portfolio_vol_resample_freq,
                                                                          br.portfolio_vol_resample_type,
                                                                          period_shift=br.portfolio_vol_period_shift)

            # portfolio, portfolio_leverage_df = risk_engine.calculate_vol_adjusted_returns(portfolio, br = br)

        # multiply portfolio leverage * individual signals to get final position signals
        length_cols = len(signal_df.columns)
        leverage_matrix = numpy.transpose(
            numpy.repeat(portfolio_leverage_df.values.flatten()[numpy.newaxis, :], length_cols, 0))

        # final portfolio signals (including signal & portfolio leverage)
        portfolio_signal = pandas.DataFrame(
            data=numpy.multiply(leverage_matrix, signal_df.values),
            index=signal_df.index, columns=signal_df.columns)

        # later when we plot the portfolio components, we do that without weighting the individual components
        portfolio_signal_before_weighting = portfolio_signal.copy()

        if br.portfolio_combination is not None:
            if 'sum' in br.portfolio_combination:
                pass
            elif br.portfolio_combination == 'mean' \
                    or (br.portfolio_combination == 'weighted' and isinstance(br.portfolio_combination_weights, dict)):
                portfolio_signal = pandas.DataFrame(data=(portfolio_signal.values * adjusted_weights_matrix),
                                                    index=portfolio_signal.index,
                                                    columns=portfolio_signal.columns)
        else:
            portfolio_signal = pandas.DataFrame(data=(portfolio_signal.values * adjusted_weights_matrix),
                                                index=portfolio_signal.index,
                                                columns=portfolio_signal.columns)

        portfolio_total_longs, portfolio_total_shorts, portfolio_net_exposure, portfolio_total_exposure \
            = self.calculate_exposures(portfolio_signal)

        # apply position limits?
        position_clip_adjustment = risk_engine.calculate_position_clip_adjustment \
            (portfolio_net_exposure, portfolio_total_exposure, br)

        # if we have any position clip adjustment, for example related to max position sizes
        if position_clip_adjustment is not None:
            position_clip_adjustment_matrix = numpy.transpose(
                numpy.repeat(position_clip_adjustment.values.flatten()[numpy.newaxis, :], length_cols, 0))

            # recalculate portfolio signals after adjustment (for individual components - without
            # weighting each signal separately)
            portfolio_signal_before_weighting = pandas.DataFrame(
                data=(portfolio_signal_before_weighting.values * position_clip_adjustment_matrix),
                index=portfolio_signal_before_weighting.index,
                columns=portfolio_signal_before_weighting.columns)

            # recalculate portfolio signal after adjustment (for portfolio level positions)
            portfolio_signal = pandas.DataFrame(
                data=(portfolio_signal.values * position_clip_adjustment_matrix),
                index=portfolio_signal.index,
                columns=portfolio_signal.columns)

            # recalculate portfolio leverage with position constraint (multiply vectors elementwise)
            portfolio_leverage_df = pandas.DataFrame(
                data=(portfolio_leverage_df.values * position_clip_adjustment.values),
                index=portfolio_leverage_df.index,
                columns=portfolio_leverage_df.columns)

            # recalculate total long, short, net and absolute exposures of the whole portfolio after the position
            # clip adjustment
            portfolio_total_longs, portfolio_total_shorts, portfolio_net_exposure, portfolio_total_exposure \
                = self.calculate_exposures(portfolio_signal)

        # calculate final portfolio returns with the amended portfolio leverage (by default just 1s)
        portfolio = calculations.calculate_signal_returns_with_tc_matrix(portfolio_leverage_df, portfolio, tc=tc)

        # assign all the property variables
        self._signal = signal_df  # individual signals (before portfolio leverage)
        self._portfolio_signal = portfolio_signal  # individual signals (AFTER portfolio leverage/constraints)
        self._portfolio_leverage = portfolio_leverage_df  # leverage on portfolio

        self._portfolio = portfolio

        # calculate each period of trades
        self._portfolio_trade = self._portfolio_signal - self._portfolio_signal.shift(1)

        # expressing trades/positions in terms of notionals
        self._portfolio_signal_notional = None
        self._portfolio_signal_trade_notional = None

        # expressing trades/positions in terms of contracts (useful for futures)
        self._portfolio_signal_contracts = None
        self._portfolio_signal_trade_contracts = None

        self._portfolio_total_longs = portfolio_total_longs
        self._portfolio_total_shorts = portfolio_total_shorts
        self._portfolio_net_exposure = portfolio_net_exposure
        self._portfolio_total_exposure = portfolio_total_exposure

        self._portfolio_total_longs_notional = None
        self._portfolio_total_shorts_notional = None
        self._portfolio_net_exposure_notional = None
        self._portfolio_total_exposure_notional = None
        self._portfolio_signal_trade_notional_sizes = None

        # also create other measures of portfolio
        # portfolio & trades in terms of a predefined notional (in USD)
        # portfolio & trades in terms of contract sizes (particularly useful for futures)
        if br.portfolio_notional_size is not None:
            # express positions in terms of the notional size specified
            self._portfolio_signal_notional = self._portfolio_signal * br.portfolio_notional_size
            self._portfolio_signal_trade_notional = self._portfolio_signal_notional - self._portfolio_signal_notional.shift(
                1)

            df_trades_sizes = pandas.DataFrame()

            rounded_portfolio_signal_trade_notional = self._portfolio_signal_trade_notional.round(2)

            for k in rounded_portfolio_signal_trade_notional.columns:
                df_trades_sizes[k] = pandas.value_counts(rounded_portfolio_signal_trade_notional[k], sort=True)

            df_trades_sizes = df_trades_sizes[df_trades_sizes.index != 0]

            self._portfolio_signal_trade_notional_sizes = df_trades_sizes

            self._portfolio_total_longs_notional = portfolio_total_longs * br.portfolio_notional_size
            self._portfolio_total_shorts_notional = portfolio_total_shorts * br.portfolio_notional_size
            self._portfolio_net_exposure_notional = portfolio_net_exposure * br.portfolio_notional_size
            self._portfolio_total_exposure_notional = portfolio_total_exposure * br.portfolio_notional_size

            # get the positions in terms of the contract sizes
            notional_copy = self._portfolio_signal_notional.copy(deep=True)
            notional_copy_cols = [x.split('.')[0] for x in notional_copy.columns]
            notional_copy_cols = [x + '.contract-value' for x in notional_copy_cols]

            notional_copy.columns = notional_copy_cols

            # can only give contract sizes if these are defined
            if contract_value_df is not None:
                contract_value_df = contract_value_df[notional_copy_cols]
                notional_df, contract_value_df = notional_copy.align(contract_value_df, join='left', axis='index')

                # careful make sure orders of magnitude are same for the notional and the contract value
                self._portfolio_signal_contracts = notional_df / contract_value_df
                self._portfolio_signal_contracts.columns = self._portfolio_signal_notional.columns
                self._portfolio_signal_trade_contracts = self._portfolio_signal_contracts - self._portfolio_signal_contracts.shift(
                    1)

        self._pnl = _pnl  # individual signals P&L (before portfolio volatility targeting, position limits etc)

        # _pnl_components of individual assets after all the portfolio level risk signals and position limits have been applied
        self._components_pnl = calculations.calculate_signal_returns_with_tc_matrix(portfolio_signal_before_weighting,
                                                                                    returns_df, tc=tc)
        self._components_pnl.columns = pnl_cols

        # TODO FIX very slow - hence only calculate on demand
        # _pnl_trades = calculations.calculate_individual_trade_gains(signal_df, _pnl)
        self._pnl_trades = None
        self._components_pnl_trades = None

        self._trade_no = None
        self._portfolio_trade_no = None

        self._portfolio.columns = ['Port']

        self._pnl_ret_stats = RetStats(self._pnl, br.ann_factor)
        self._components_pnl_ret_stats = RetStats(self._components_pnl, br.ann_factor)
        self._portfolio_ret_stats = RetStats(self._portfolio, br.ann_factor)

        # TODO parallel version still work in progress!

        logger.info("Cumulative index calculations")

        if False: # market_constants.backtest_thread_no[market_constants.generic_plat] > 1 and run_in_parallel:
            swim_pool = SwimPool(multiprocessing_library=market_constants.multiprocessing_library)

            pool = swim_pool.create_pool(thread_technique=market_constants.backtest_thread_technique,
                                         thread_no=market_constants.backtest_thread_no[market_constants.generic_plat])

            r1 = pool.apply_async(self._pnl_ret_stats.calculate_ret_stats)
            r2 = pool.apply_async(self._components_pnl_ret_stats.calculate_ret_stats)
            r3 = pool.apply_async(self._portfolio_ret_stats.calculate_ret_stats)

            resultsA = pool.apply_async(calculations.create_mult_index, args=(self._pnl,))
            resultsB = pool.apply_async(calculations.create_mult_index, args=(self._components_pnl,))
            resultsC = pool.apply_async(calculations.create_mult_index, args=(self._portfolio,))

            swim_pool.close_pool(pool)

            self._pnl_ret_stats = r1.get()
            self._components_pnl_ret_stats = r2.get()
            self._portfolio_ret_stats = r3.get()

            self._pnl_cum = resultsA.get()
            self._components_pnl_cum = resultsB.get()
            self._portfolio_cum = resultsC.get()

        else:
            # calculate return statistics of the each asset/signal after signal leverage (but before portfolio level constraints)
            # self._ret_stats_pnl.calculate_ret_stats()

            # calculate return statistics of the each asset/signal after signal leverage AND after portfolio level constraints
            # self._ret_stats_pnl_components.calculate_ret_stats()

            # calculate return statistics of the final portfolio
            # self._ret_stats_portfolio.calculate_ret_stats()

            # calculate individual signals cumulative P&L after signal leverage but before portfolio level constraints
            self._pnl_cum = calculations.create_mult_index(self._pnl)

            # calculate individual signals cumulative P&L after signal leverage AND after portfolio level constraints
            self._components_pnl_cum = calculations.create_mult_index(self._components_pnl)

            # calculate final portfolio cumulative P&L
            self._portfolio_cum = calculations.create_mult_index(self._portfolio)  # portfolio cumulative P&L

        logger.info("Completed cumulative index calculations")

        self._pnl_cum.columns = pnl_cols
        self._components_pnl_cum.columns = pnl_cols
        self._portfolio_cum.columns = ['Port']

    def calculate_exposures(self, portfolio_signal):
        """Calculates time series for the total longs, short, net and absolute exposure on an aggregated portfolio basis.

        Parameters
        ----------
        portfolio_signal : DataFrame
            Signals for each asset in the portfolio after all weighting, portfolio & signal level volatility adjustments

        Returns
        -------
        DataFrame (list)

        """
        # calculate total portfolio longs/total portfolio shorts/total portfolio exposure
        portfolio_total_longs = pandas.DataFrame(portfolio_signal[portfolio_signal > 0].sum(axis=1))
        portfolio_total_shorts = pandas.DataFrame(portfolio_signal[portfolio_signal < 0].sum(axis=1))

        portfolio_total_longs.columns = ['Total Longs']
        portfolio_total_shorts.columns = ['Total Shorts']

        # NOTE: careful usage of signs (portfolio_total_shorts are NEGATIVE)
        portfolio_net_exposure = pandas.DataFrame(
            index=portfolio_signal.index, columns=['Net Exposure'],
            data=portfolio_total_longs.values + portfolio_total_shorts.values)

        portfolio_total_exposure = pandas.DataFrame(
            index=portfolio_signal.index, columns=['Total Exposure'],
            data=portfolio_total_longs.values - portfolio_total_shorts.values)

        return portfolio_total_longs, portfolio_total_shorts, portfolio_net_exposure, portfolio_total_exposure

    def create_portfolio_weights(self, br, _pnl, method='mean'):
        """Calculates P&L of a trading strategy and statistics to be retrieved later

        Parameters
        ----------
        br : BacktestRequest
            Parameters for the backtest specifying start date, finish data, transaction costs etc.

        _pnl : pandas.DataFrame
            Contains the daily P&L for the portfolio

        method : String
            'mean' - assumes equal weighting for each signal
            'weighted' - can use predefined user weights (eg. if we assign weighting of 1, 1, 0.5, for three signals
            the third signal will have a weighting of half versus the others)
        
        weights : dict
            Portfolio weights

        Returns
        -------
        pandas.DataFrame
            Contains the portfolio weights
        """
        if method == 'mean':
            weights_vector = numpy.ones(len(_pnl.columns))
        elif method == 'weighted' or 'weighted-sum':
            # get the weights for each asset
            weights_vector = numpy.array([float(br.portfolio_combination_weights[col]) for col in _pnl.columns])

        # repeat this down for every day
        weights_matrix = numpy.repeat(weights_vector[numpy.newaxis, :], len(_pnl.index), 0)

        # where we don't have old price data, make the weights 0 there
        ind = numpy.isnan(_pnl.values)
        weights_matrix[ind] = 0

        if method != 'weighted-sum':
            # the total weights will vary, as historically might not have all the assets trading
            total_weights = numpy.sum(weights_matrix, axis=1)

            # replicate across columns
            total_weights = numpy.transpose(numpy.repeat(total_weights[numpy.newaxis, :], len(_pnl.columns), 0))

            # to avoid divide by zero
            total_weights[total_weights == 0.0] = 1.0

            adjusted_weights_matrix = weights_matrix / total_weights
            adjusted_weights_matrix[ind] = numpy.nan

            return adjusted_weights_matrix

        return weights_matrix

    def backtest_output(self):
        return

    ### Get PnL of individual assets before portfolio constraints
    def pnl(self):
        """Gets P&L returns of all the individual sub_components of the model (before any portfolio level leverage is applied)

        Returns
        -------
        pandas.Dataframe
        """
        return self._pnl

    def trade_no(self):
        """Gets number of trades for each signal in the backtest (before

        Returns
        -------
        pandas.Dataframe
        """

        if self._trade_no is None:
            calculations = Calculations()
            self._trade_no = calculations.calculate_trade_no(self._signal)

        return self._trade_no

    def pnl_trades(self):
        """Gets P&L of each individual trade per signal

        Returns
        -------
        pandas.Dataframe
        """

        if self._pnl_trades is None:
            calculations = Calculations()
            self._pnl_trades = calculations.calculate_individual_trade_gains(self._signal, self._pnl)

        return self._pnl_trades

    def pnl_desc(self):
        """Gets P&L return statistics in a string format

        Returns
        -------
        str
        """
        return self._ret_stats_signals.summary()

    def pnl_ret_stats(self):
        """Gets P&L return statistics of individual strategies as class to be queried

        Returns
        -------
        TimeSeriesDesc
        """

        return self._pnl_ret_stats

    def pnl_cum(self):
        """Gets P&L as a cumulative time series of individual assets

        Returns
        -------
        pandas.DataFrame
        """

        return self._pnl_cum

    ### Get PnL of individual assets AFTER portfolio constraints
    def components_pnl(self):
        """Gets P&L returns of all the individual subcomponents of the model (after portfolio level leverage is applied)

        Returns
        -------
        pandas.Dataframe
        """
        return self._components_pnl

    def components_pnl_trades(self):
        """Gets P&L of each individual trade per signal

        Returns
        -------
        pandas.Dataframe
        """

        if self._components_pnl_trades is None:
            calculations = Calculations()
            self._components_pnl_trades = calculations.calculate_individual_trade_gains(self._signal,
                                                                                        self._components_pnl)

        return self._components_pnl_trades

    def components_pnl_desc(self):
        """Gets P&L of individual components as return statistics in a string format

        Returns
        -------
        str
        """
        # return self._ret_stats_signals.summary()

    def components_pnl_ret_stats(self):
        """Gets P&L return statistics of individual strategies as class to be queried

        Returns
        -------
        TimeSeriesDesc
        """

        return self._components_pnl_ret_stats

    def components_pnl_cum(self):
        """Gets P&L as a cumulative time series of individual assets (after portfolio level leverage adjustments)

        Returns
        -------
        pandas.DataFrame
        """

        return self._components_pnl_cum

    ### Get PnL of the final portfolio

    def portfolio_cum(self):
        """Gets P&L as a cumulative time series of portfolio

        Returns
        -------
        pandas.DataFrame
        """

        return self._portfolio_cum

    def portfolio_pnl(self):
        """Gets portfolio returns in raw form (ie. not indexed into cumulative form)

        Returns
        -------
        pandas.DataFrame
        """

        return self._portfolio

    def portfolio_pnl_desc(self):
        """Gets P&L return statistics of portfolio as string

        Returns
        -------
        pandas.DataFrame
        """

        return self._portfolio_ret_stats.summary()

    def portfolio_pnl_ret_stats(self):
        """Gets P&L return statistics of portfolio as class to be queried

        Returns
        -------
        RetStats
        """

        return self._portfolio_ret_stats

    def individual_leverage(self):
        """Gets leverage for each asset historically

        Returns
        -------
        pandas.DataFrame
        """

        return self._individual_leverage

    def portfolio_leverage(self):
        """Gets the leverage for the portfolio

        Returns
        -------
        pandas.DataFrame
        """

        return self._portfolio_leverage

    def portfolio_trade_no(self):
        """Gets number of trades for each signal in the backtest (after both signal and portfolio level vol adjustment)

        Returns
        -------
        pandas.Dataframe
        """

        if self._portfolio_trade_no is None:
            calculations = Calculations()
            self._portfolio_trade_no = calculations.calculate_trade_no(self._portfolio_signal)

        return self._portfolio_trade_no

    def portfolio_signal(self):
        """Gets the signals (with individual leverage & portfolio leverage) for each asset, which
        equates to what we would trade in practice

        Returns
        -------
        DataFrame
        """

        return self._portfolio_signal

    def portfolio_total_longs(self):
        """Gets the total long exposure in the portfolio

        Returns
        -------
        DataFrame
        """

        return self._portfolio_total_longs

    def portfolio_total_shorts(self):
        """Gets the total short exposure in the portfolio

        Returns
        -------
        DataFrame
        """

        return self._portfolio_total_shorts

    def portfolio_net_exposure(self):
        """Gets the total net exposure of the portfolio

        Returns
        -------
        DataFrame
        """

        return self._portfolio_net_exposure

    def portfolio_total_exposure(self):
        """Gets the total absolute exposure of the portfolio

        Returns
        -------
        DataFrame
        """

        return self._portfolio_total_exposure

    def portfolio_total_longs_notional(self):
        """Gets the total long exposure in the portfolio scaled by notional

        Returns
        -------
        DataFrame
        """

        return self._portfolio_total_longs_notional

    def portfolio_total_shorts_notional(self):
        """Gets the total short exposure in the portfolio scaled by notional

        Returns
        -------
        DataFrame
        """

        return self._portfolio_total_shorts_notional

    def portfolio_net_exposure_notional(self):
        """Gets the total net exposure of the portfolio scaled by notional

        Returns
        -------
        DataFrame
        """

        return self._portfolio_net_exposure_notional

    def portfolio_total_exposure_notional(self):
        """Gets the total absolute exposure of the portfolio scaled by notional

        Returns
        -------
        DataFrame
        """

        return self._portfolio_total_exposure_notional

    def portfolio_trade(self):
        """Gets the trades (with individual leverage & portfolio leverage) for each asset, which
        we'd need to execute

        Returns
        -------
        DataFrame
        """

        return self._portfolio_trade

    def portfolio_signal_notional(self):
        """Gets the signals (with individual leverage & portfolio leverage) for each asset, which
        equates to what we would have a positions in practice, scaled by a notional amount we have already specified

        Returns
        -------
        DataFrame
        """

        return self._portfolio_signal_notional

    def portfolio_trade_notional(self):
        """Gets the trades (with individual leverage & portfolio leverage) for each asset, which
        we'd need to execute, scaled by a notional amount we have already specified

        Returns
        -------
        DataFrame
        """

        return self._portfolio_signal_trade_notional

    def portfolio_trade_notional_sizes(self):
        """Gets the number of trades (with individual leverage & portfolio leverage) for each asset, which
        we'd need to execute, scaled by a notional amount we have already specified

        Returns
        -------
        DataFrame
        """

        return self._portfolio_signal_trade_notional_sizes

    def portfolio_signal_contracts(self):
        """Gets the signals (with individual leverage & portfolio leverage) for each asset, which
        equates to what we would have a positions in practice, scaled by a notional amount and into contract sizes (eg. for futures)
        which we need to specify in another dataframe

        Returns
        -------
        DataFrame
        """

        return self._portfolio_signal_contracts

    def portfolio_trade_contracts(self):
        """Gets the trades (with individual leverage & portfolio leverage) for each asset, which
        we'd need to execute, scaled by a notional amount we have already specified and into contract sizes (eg. for futures)
        which we need to specify in another dataframe

        Returns
        -------
        DataFrame
        """

        return self._portfolio_signal_trade_contracts

    def signal(self):
        """Gets signal for each asset (with individual leverage, but excluding portfolio leverage constraints) for each asset

        Returns
        -------
        pandas.DataFrame
        """

        return self._signal


########################################################################################################################

import abc
import pandas
import datetime
import copy

from chartpy import Chart, Style, ChartConstants

from finmarketpy.economics import TechParams
from findatapy.timeseries import Calculations, RetStats, Filter


class TradingModel(object):
    """Abstract class which wraps around Backtest, providing convenient functions for analaysis. Implement your own
    subclasses of this for your own strategy. See tradingmodelfxtrend_example.py for a simple implementation of a
    FX trend following strategy.

    """

    #### Default parameters for outputting of results from trading model
    SAVE_FIGURES = True
    SHOW_CHARTS = True
    SHOW_TITLES = True

    DEFAULT_PLOT_ENGINE = ChartConstants().chartfactory_default_engine
    SCALE_FACTOR = ChartConstants().chartfactory_scale_factor
    CHART_SOURCE = ChartConstants().chartfactory_source
    CHART_STYLE = Style()

    DUMP_CSV = ''
    DUMP_PATH = datetime.date.today().strftime("%Y%m%d") + ' '

    # logger = LoggerManager().getLogger(__name__)

    def __init__(self):
        pass

    # to be implemented by every trading strategy
    @abc.abstractmethod
    def load_parameters(self, br=None):
        """Fills parameters for the backtest, such as start-end dates, transaction costs etc. To
        be implemented by subclass. Can overwrite it with our own BacktestRequest.
        """
        return

    @abc.abstractmethod
    def load_assets(self, br=None):
        """Loads time series for the assets to be traded and also for data for generating signals. Can overwrite it
        with our own BacktestRequest.
        """
        return

    @abc.abstractmethod
    def construct_signal(self, spot_df, spot_df2, tech_params, br, run_in_parallel=False):
        """Constructs signal from pre-loaded time series

        Parameters
        ----------
        spot_df : pandas.DataFrame
            Market time series for generating signals

        spot_df2 : pandas.DataFrame
            Market time series for generated signals (can be of different frequency)

        tech_params : TechParams
            Parameters for generating signals

        run_in_parallel : bool
            Allow signal calculation in parallel
        """
        return

    ####### Generic functions for every backtest
    def construct_strategy(self, br=None, run_in_parallel=False):
        """Constructs the returns for all the strategies which have been specified.

        It gets backtesting parameters from fill_backtest_request (although these can be overwritten
        and then market data from fill_assets

        Parameters
        ----------
        br : BacktestRequest
            Parameters which define the backtest (for example start date, end date, transaction costs etc.

        """

        logger = LoggerManager().getLogger(__name__)

        # get the parameters for backtesting
        if hasattr(self, 'br'):
            br = self.br
        elif br is None:
            br = self.load_parameters()

        # get market data for backtest
        market_data = self.load_assets()

        asset_df = market_data[0]
        spot_df = market_data[1]
        spot_df2 = market_data[2]
        basket_dict = market_data[3]
        # contract_value_df = market_data[4]

        # optional database output
        contract_value_df = None

        if len(market_data) == 5:
            contract_value_df = market_data[4]

        if hasattr(br, 'tech_params'):
            tech_params = br.tech_params
        else:
            tech_params = TechParams()

        cum_results = pandas.DataFrame(index=asset_df.index)
        port_leverage = pandas.DataFrame(index=asset_df.index)

        from collections import OrderedDict
        ret_stats_results = OrderedDict()

        bask_results = {}

        bask_keys = basket_dict.keys()

        # each portfolio key calculate returns - can put parts of the portfolio in the key
        if market_constants.backtest_thread_no[market_constants.generic_plat] > 1 and run_in_parallel:
            swim_pool = SwimPool(multiprocessing_library=market_constants.multiprocessing_library)

            pool = swim_pool.create_pool(thread_technique=market_constants.backtest_thread_technique,
                                         thread_no=market_constants.backtest_thread_no[market_constants.generic_plat])

            mult_results = []

            #start = asset_df.index[0]
            #finish = asset_df.index[-1]

            # calculate sub substrategies in sub-processes
            # TODO cut up in time chunks
            for key in bask_keys:

                if key != self.FINAL_STRATEGY:
                    logger.info("Calculating (parallel) " + key)

                    asset_cut_df = asset_df[[x + '.close' for x in basket_dict[key]]]
                    spot_cut_df = spot_df[[x + '.close' for x in basket_dict[key]]]

                    mult_results.append(
                        pool.apply_async(self.construct_individual_strategy, args=(br, spot_cut_df, spot_df2, asset_cut_df,
                                                                                   tech_params, key, contract_value_df,
                                                                                   False, True,))
                    )

                    # calculate final strategy separately in my main process (so don't have issues with pickling back large output)

            logger.info("Calculating final strategy " + self.FINAL_STRATEGY)

            # calculate the final strategy separately (can often be a lot larger)
            asset_cut_df = asset_df[[x + '.close' for x in basket_dict[self.FINAL_STRATEGY]]]
            spot_cut_df = spot_df[[x + '.close' for x in basket_dict[self.FINAL_STRATEGY]]]

            desc, results, leverage, stats, key, backtest = \
                    self.construct_individual_strategy(br, spot_cut_df, spot_df2,
                                                           asset_cut_df,
                                                           tech_params,
                                                           self.FINAL_STRATEGY,
                                                           contract_value_df, True,
                                                           False)

            results.columns = desc

            cum_results[results.columns[0]] = results
            port_leverage[results.columns[0]] = leverage
            ret_stats_results[key] = stats

            self._assign_final_strategy_results(results, backtest)

            for p in mult_results:

                desc, results, leverage, stats, key, backtest = p.get()

                results.columns = desc

                cum_results[results.columns[0]] = results
                port_leverage[results.columns[0]] = leverage
                ret_stats_results[key] = stats

                if key == self.FINAL_STRATEGY:
                    self._assign_final_strategy_results(results, backtest)

            try:
                swim_pool.close_pool(pool)
            except:
                pass

        else:
            for key in bask_keys:
                logger.info("Calculating (single thread) " + key)

                asset_cut_df = asset_df[[x + '.close' for x in basket_dict[key]]]
                spot_cut_df = spot_df[[x + '.close' for x in basket_dict[key]]]

                desc, results, leverage, stats, key, backtest = \
                    self.construct_individual_strategy(br, spot_cut_df, spot_df2, asset_cut_df,
                                                                       tech_params, key,
                                                                       contract_value_df, False, False)

                #results = backtest.portfolio_cum()
                results.columns = desc

                cum_results[results.columns[0]] = results
                port_leverage[results.columns[0]] = leverage
                ret_stats_results[key] = stats

                if key == self.FINAL_STRATEGY:
                    self._assign_final_strategy_results(results, backtest)

        # get benchmark for comparison
        benchmark = self.construct_strategy_benchmark()

        cum_results_benchmark = self.compare_strategy_vs_benchmark(br, cum_results, benchmark)

        self._strategy_group_benchmark_pnl_ret_stats = ret_stats_results

        if hasattr(self, '_benchmark_ret_stats'):
            ret_stats_list = ret_stats_results
            ret_stats_list['Benchmark'] = (self._strategy_benchmark_pnl_ret_stats)
            self._strategy_group_benchmark_pnl_ret_stats = ret_stats_list

        self._strategy_group_pnl = cum_results  # individual parts (all after individually applying portfolio level vol adjustment)
        self._strategy_group_pnl_ret_stats = ret_stats_results

        self._strategy_group_leverage = port_leverage
        self._strategy_group_benchmark_pnl = cum_results_benchmark

    def _assign_final_strategy_results(self, results, backtest):
            # for a key, designated as the final strategy save that as the "strategy"

            self._strategy_pnl = results  # cumulative P&L for the final strategy
            self._strategy_components_pnl = backtest.pnl_cum()  # P&L of the individual components (after portfolio level vol adjustments)
            self._strategy_components_pnl_ret_stats = backtest.components_pnl_ret_stats().split_into_dict()  # backtest.get_pnl_components_ret_stats()

            self._strategy_pnl_ret_stats = backtest.portfolio_pnl_ret_stats()
            self._strategy_leverage = backtest.portfolio_leverage()

            # collect the position sizes and trade sizes (in several different formats)
            self._strategy_signal = backtest.portfolio_signal()
            self._strategy_trade_no = backtest.portfolio_trade_no()
            self._strategy_trade = backtest.portfolio_trade()

            # scaled by notional
            self._strategy_signal_notional = backtest.portfolio_signal_notional()
            self._strategy_trade_notional = backtest.portfolio_trade_notional()
            self._strategy_trade_notional_sizes = backtest.portfolio_trade_notional_sizes()

            # scaled by notional and adjusted into contract sizes
            self._strategy_signal_contracts = backtest.portfolio_signal_contracts()
            self._strategy_trade_contracts = backtest.portfolio_trade_contracts()

            self._strategy_group_pnl_trades = backtest.pnl_trades()  # get individual trades P&L before (before portfolio adjustment)
            self._strategy_pnl_trades_components = backtest.components_pnl_trades()

            self._strategy_total_longs = backtest.portfolio_total_longs()
            self._strategy_total_shorts = backtest.portfolio_total_shorts()
            self._strategy_net_exposure = backtest.portfolio_net_exposure()
            self._strategy_total_exposure = backtest.portfolio_total_exposure()

            self._strategy_total_longs_notional = backtest.portfolio_total_longs_notional()
            self._strategy_total_shorts_notional = backtest.portfolio_total_shorts_notional()
            self._strategy_net_exposure_notional = backtest.portfolio_net_exposure_notional()
            self._strategy_total_exposure_notional = backtest.portfolio_total_exposure_notional()

    def construct_individual_strategy(self, br, spot_df, spot_df2, asset_df, tech_params, key, contract_value_df,
                                      run_in_parallel, compress_output):
        """Combines the signal with asset returns to find the returns of an individual strategy

        Parameters
        ----------
        br : BacktestRequest
            Parameters for backtest such as start and finish dates

        spot_df : pandas.DataFrame
            Market time series for generating signals

        spot_df2 : pandas.DataFrame
            Secondary Market time series for generated signals (can be of different frequency)

        tech_params : TechParams
            Parameters for generating signals

        contract_value_df : pandas.DataFrame
            Dataframe with the contract sizes for each asset

        Returns
        -------
        portfolio_cum : pandas.DataFrame
        backtest : Backtest
        """
        backtest = Backtest()

        logger = LoggerManager().getLogger(__name__)

        logger.info("Calculating trading signals for " + key + "...")

        signal = self.construct_signal(spot_df, spot_df2, tech_params, br, run_in_parallel=run_in_parallel)

        logger.info("Calculated trading signals for " + key)

        backtest.calculate_trading_PnL(br, asset_df, signal,
                                       contract_value_df,
                                       run_in_parallel)  # calculate P&L (and adjust signals for vol etc)

        if br.write_csv: backtest.pnl_cum().to_csv(self.DUMP_CSV + key + ".csv")

        if br.calc_stats:
            desc = [key + ' ' + str(backtest.portfolio_pnl_desc()[0])]
        else:
            desc = [key]

        # for final strategy return heavyweight backtest object (contains lots of auxilliary information about trades etc)
        if key == self.FINAL_STRATEGY and compress_output:
            logger.debug('Compressing ' + key)

            backtest = blosc.compress(pickle.dumps(backtest))

            logger.debug('Compressed ' + key)

        logger.info('Calculated for ' + key)

        if key != self.FINAL_STRATEGY:
            return desc, backtest.portfolio_cum(), backtest.portfolio_leverage(), backtest.pnl_ret_stats(), key, None

        return desc, backtest.portfolio_cum(), backtest.portfolio_leverage(), backtest.pnl_ret_stats(), key, backtest

        # have lightweight output for subcomponents of portfolio
        # return desc, backtest.portfolio_cum(), backtest.portfolio_leverage(), backtest.portfolio_pnl_ret_stats(), \
        #       key, _

    def compare_strategy_vs_benchmark(self, br, strategy_df, benchmark_df):
        """Compares the trading strategy we are backtesting against a benchmark

        Parameters
        ----------
        br : BacktestRequest
            Parameters for backtest such as start and finish dates
        strategy_df : pandas.DataFrame
            Strategy time series
        benchmark_df : pandas.DataFrame
            Benchmark time series
        """

        if br.include_benchmark:
            ret_stats = RetStats()
            risk_engine = RiskEngine()
            filter = Filter()
            calculations = Calculations()

            # align strategy time series with that of benchmark
            benchmark_df.columns = [x + ' be' for x in benchmark_df.columns]
            strategy_df, benchmark_df = strategy_df.align(benchmark_df, join='left', axis=0)

            # if necessary apply vol target to benchmark (to make it comparable with strategy)
            if br.portfolio_vol_adjust is True:
                benchmark_df = risk_engine.calculate_vol_adjusted_index_from_prices(benchmark_df, br=br)

            # only calculate return statistics if this has been specified (note when different frequencies of data
            # might underrepresent vol
            # if calc_stats:
            benchmark_df = benchmark_df.fillna(method='ffill')
            ret_stats.calculate_ret_stats_from_prices(benchmark_df, br.ann_factor)

            if br.calc_stats:
                benchmark_df.columns = ret_stats.summary()

            # realign strategy & benchmark
            strategy_benchmark_df = strategy_df.join(benchmark_df, how='inner')
            strategy_benchmark_df = strategy_benchmark_df.fillna(method='ffill')

            if br.plot_start is not None:
                strategy_benchmark_df = filter.filter_time_series_by_date(br.plot_start, br.finish_date,
                                                                          strategy_benchmark_df)

            strategy_benchmark_df = calculations.create_mult_index_from_prices(strategy_benchmark_df)

            self._strategy_benchmark_pnl = benchmark_df
            self._strategy_benchmark_pnl_ret_stats = ret_stats

            return strategy_benchmark_df

        return strategy_df

    def _flatten_list(self, list_of_lists):
        """Flattens list, particularly useful for combining baskets

        Parameters
        ----------
        list_of_lists : str (list)
            List to be flattened

        Returns
        -------

        """
        result = []

        for i in list_of_lists:
            # only append if i is a basestring (superclass of string)
            if isinstance(i, str):
                result.append(i)
            # otherwise call this function recursively
            else:
                result.extend(self._flatten_list(i))

        return result

    def strategy_name(self):
        return self.FINAL_STRATEGY

    def individual_leverage(self):
        return self._individual_leverage

    def strategy_group_pnl_trades(self):
        return self._strategy_group_pnl_trades

    ### components
    def strategy_components_pnl(self):
        return self._strategy_components_pnl

    def strategy_components_pnl_ret_stats(self):
        return self._strategy_components_pnl_ret_stats

    ### final strategy
    def strategy_pnl(self):
        return self._strategy_pnl

    def strategy_pnl_ret_stats(self):
        return self._strategy_pnl_ret_stats

    def strategy_leverage(self):
        return self._strategy_leverage

    ### final PNL strategy + benchmark
    def strategy_benchmark_pnl(self):
        return self._strategy_benchmark_pnl

    def strategy_benchmark_pnl_ret_stats(self):
        return self._strategy_benchmark_pnl_ret_stats

    ### final PNL + group
    def strategy_group_pnl(self):
        return self._strategy_group_pnl

    def strategy_group_pnl_ret_stats(self):
        return self._strategy_group_pnl_ret_stats

    ### final P&L + group + benchmark
    def strategy_group_benchmark_pnl(self):
        return self._strategy_group_benchmark_pnl

    def strategy_group_benchmark_pnl_ret_stats(self):
        return self._strategy_group_benchmark_pnl_ret_stats

    def strategy_group_leverage(self):
        return self._strategy_group_leverage

    # signals
    def strategy_signal(self):
        return self._strategy_signal

    def strategy_trade(self):
        return self._strategy_trade

    def strategy_signal_notional(self):
        return self._strategy_signal_notional

    def strategy_trade_notional(self):
        return self._strategy_trade_notional

    def strategy_trade_notional_sizes(self):
        return self._strategy_trade_notional_sizes

    def strategy_signal_contracts(self):
        return self._strategy_signal_contracts

    def strategy_trade_contracts(self):
        return self._strategy_trade_contracts

    def strategy_total_longs(self):
        return self._strategy_total_longs

    def strategy_total_shorts(self):
        return self._strategy_total_shorts

    def strategy_net_exposure(self):
        return self._strategy_net_exposure

    def strategy_total_exposure(self):
        return self._strategy_total_exposure

    def strategy_total_longs_notional(self):
        return self._strategy_total_longs_notional

    def strategy_total_shorts_notional(self):
        return self._strategy_total_shorts_notional

    def strategy_net_exposure_notional(self):
        return self._strategy_net_exposure_notional

    def strategy_total_exposure_notional(self):
        return self._strategy_total_exposure_notional

    #### Plotting

    def _reduce_plot(self, data_frame, reduce_plot=True, resample='B'):
        """Reduces the frequency of a time series to every business day so it can be plotted more easily

        Parameters
        ----------
        data_frame: pandas.DataFrame
            Strategy time series

        Returns
        -------
        pandas.DataFrame
        """
        try:
            if reduce_plot:
                # make plots on every business day (will downsample intraday data)
                data_frame = data_frame.resample(resample).last()
                data_frame = data_frame.fillna(method='pad')

            return data_frame
        except:
            return data_frame

    ##### Quick helper functions to plot aspects of the strategy such as P&L, leverage etc.
    def plot_individual_leverage(self, strip, silent_plot=False, reduce_plot=True, resample='B'):

        style = self._create_style("Leverage", "Individual Leverage", reduce_plot=reduce_plot)

        try:
            chart = Chart(self._strip_dataframe(self._reduce_plot(self._individual_leverage,
                                                                  reduce_plot=reduce_plot, resample=resample), strip),
                          engine=self.DEFAULT_PLOT_ENGINE, chart_type='line', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_group_pnl_trades(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        style = self._create_style("(bp)", "Individual Trade PnL", reduce_plot=reduce_plot)

        # zero when there isn't a trade exit
        # strategy_pnl_trades = self._strategy_pnl_trades * 100 * 100
        # strategy_pnl_trades = strategy_pnl_trades.dropna()

        # note only works with single large basket trade
        try:
            strategy_pnl_trades = self._strategy_group_pnl_trades.fillna(0) * 100 * 100
            chart = Chart(self._strip_dataframe(self._reduce_plot(strategy_pnl_trades, reduce_plot=reduce_plot,
                                                                  resample=resample), strip),
                          engine=self.DEFAULT_PLOT_ENGINE, chart_type='line', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_pnl(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        style = self._create_style("", "Strategy PnL", reduce_plot=reduce_plot)

        try:
            df = self._strip_dataframe(self._reduce_plot(self._strategy_pnl,
                                                        reduce_plot=reduce_plot, resample=resample), strip)

            chart = Chart(df, engine=self.DEFAULT_PLOT_ENGINE, chart_type='line', style=style)

            if not (silent_plot): chart.plot()

            if self.br.write_csv_pnl:
                df.to_csv(self.DUMP_PATH + self.FINAL_STRATEGY + "_pnl.csv")

            return chart
        except Exception as e:
            print(str(e))

    def plot_strategy_trade_no(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):
        df_trades = self._strategy_trade_no

        if strip is not None: df_trades.index = [k.replace(strip, '') for k in df_trades.index]

        style = self._create_style("", "", reduce_plot=reduce_plot)

        try:
            style.file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (Strategy trade no).png'
            style.html_file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (Strategy trade no).html'

            chart = Chart(self._strip_dataframe(self._reduce_plot(df_trades, reduce_plot=reduce_plot,
                                                                  resample=resample), strip),
                          engine=self.DEFAULT_PLOT_ENGINE,
                          chart_type='bar', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_signal_proportion(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        signal = self._strategy_signal

        ####### count number of long, short and flat periods in our sample
        long = signal[signal > 0].count()
        short = signal[signal < 0].count()
        flat = signal[signal == 0].count()

        df = pandas.DataFrame(index=long.index, columns=['Long', 'Short', 'Flat'])

        df['Long'] = long
        df['Short'] = short
        df['Flat'] = flat

        if strip is not None: df.index = [k.replace(strip, '') for k in df.index]

        style = self._create_style("", "")

        try:
            style.file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (Strategy signal proportion).png'
            style.html_file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (Strategy signal proportion).html'

            chart = Chart(self._strip_dataframe(self._reduce_plot(df), strip, reduce_plot=reduce_plot,
                                                resample=resample), engine=self.DEFAULT_PLOT_ENGINE,
                          chart_type='bar', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_leverage(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):
        style = self._create_style("Leverage", "Strategy Leverage", reduce_plot=reduce_plot)

        try:
            chart = Chart(self._strip_dataframe(self._reduce_plot(self._strategy_leverage,
                                                                  reduce_plot=reduce_plot, resample=resample), strip),
                          engine=self.DEFAULT_PLOT_ENGINE, chart_type='line', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    ##### plot the individual components cumulative returns and return statistics (including portfolio level constraints)
    def plot_strategy_components_pnl(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        style = self._create_style("Ind Components", "Strategy PnL Components", reduce_plot=reduce_plot)

        try:
            chart = Chart(self._strip_dataframe(self._reduce_plot(self._strategy_components_pnl,
                                                                  reduce_plot=reduce_plot, resample=resample), strip),
                          engine=self.DEFAULT_PLOT_ENGINE, chart_type='line', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_components_pnl_ir(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_components_pnl_ret_stats, 'IR', 'Ind Component',
                                           'Ind Component IR',
                                           strip=strip,
                                           silent_plot=silent_plot)

    def plot_strategy_components_pnl_returns(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_components_pnl_ret_stats, 'Returns', 'Ind Component (%)',
                                           'Ind Component Returns', strip=strip,
                                           silent_plot=silent_plot)

    def plot_strategy_components_pnl_vol(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_components_pnl_ret_stats, 'Vol', 'Ind Component (%)',
                                           'Ind Component Vol', strip=strip,
                                           silent_plot=silent_plot)

    def plot_strategy_components_pnl_drawdowns(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_components_pnl_ret_stats, 'Drawdowns', 'Ind Component (%)',
                                           'Ind Component Drawdowns', strip=strip,
                                           silent_plot=silent_plot)

    def plot_strategy_components_pnl_yoy(self, strip=None, silent_plot=False):

        return self.plot_yoy_helper(self._strategy_components_pnl_ret_stats, 'Ind Component YoY', 'Ind Component (%)',
                                    strip=strip, silent_plot=silent_plot)

    ##### plot the cumulative returns and return statistics for the strategy group
    ##### this will plot the final strategy, benchmark and all the individual baskets (as though they were run by themselves)
    def plot_strategy_group_benchmark_pnl(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):
        logger = LoggerManager().getLogger(__name__)

        style = self._create_style("", "Group Benchmark PnL - cumulative")

        strat_list = self._strategy_group_benchmark_pnl.columns  # .sort_values()

        for line in strat_list: logger.info(line)

        # plot cumulative line of returns
        chart = Chart(
            self._strip_dataframe(self._reduce_plot(self._strategy_group_benchmark_pnl,
                                                    reduce_plot=reduce_plot, resample=resample), strip),
            engine=self.DEFAULT_PLOT_ENGINE, chart_type='line', style=style)

        if not (silent_plot): chart.plot()
        return chart

    def plot_strategy_group_benchmark_pnl_ir(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_group_benchmark_pnl_ret_stats, 'IR', '',
                                           'Group Benchmark IR',
                                           strip=strip, silent_plot=silent_plot)

    def plot_strategy_group_benchmark_pnl_returns(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_group_benchmark_pnl_ret_stats, 'Returns', '(%)',
                                           'Group Benchmark Returns',
                                           strip=strip, silent_plot=silent_plot)

    def plot_strategy_group_benchmark_pnl_vol(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_group_benchmark_pnl_ret_stats, 'Vol', '(%)',
                                           'Group Benchmark Vol',
                                           strip=strip, silent_plot=silent_plot)

    def plot_strategy_group_benchmark_pnl_drawdowns(self, strip=None, silent_plot=False):
        return self._plot_ret_stats_helper(self._strategy_group_benchmark_pnl_ret_stats, 'Drawdowns', '(%)',
                                           'Group Benchmark Drawdowns', strip=strip,
                                           silent_plot=silent_plot)

    def _plot_ret_stats_helper(self, ret_stats, metric, title, file_tag, strip=None, silent_plot=False):
        style = self._create_style(title, file_tag)
        keys = ret_stats.keys()
        ret_metric = []

        for key in keys:
            if metric == 'IR':
                ret_metric.append(ret_stats[key].inforatio()[0])
            elif metric == 'Returns':
                ret_metric.append(ret_stats[key].ann_returns()[0] * 100)
            elif metric == 'Vol':
                ret_metric.append(ret_stats[key].ann_vol()[0] * 100)
            elif metric == 'Drawdowns':
                ret_metric.append(ret_stats[key].drawdowns()[0] * 100)

        if strip is not None: keys = [k.replace(strip, '') for k in keys]

        ret_stats = pandas.DataFrame(index=keys, data=ret_metric, columns=[metric])
        # ret_stats = ret_stats.sort_index()
        style.file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (' + file_tag + ') ' + str(
            style.scale_factor) + '.png'
        style.html_file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (' + file_tag + ') ' + str(
            style.scale_factor) + '.html'
        style.display_brand_label = False

        chart = Chart(ret_stats, engine=self.DEFAULT_PLOT_ENGINE, chart_type='bar', style=style)
        if not (silent_plot): chart.plot()

        return chart

    def plot_strategy_group_benchmark_pnl_yoy(self, strip=None, silent_plot=False):

        return self.plot_yoy_helper(self._strategy_group_benchmark_pnl_ret_stats, "", "Group Benchmark PnL YoY",
                                    strip=strip, silent_plot=silent_plot)

    def plot_yoy_helper(self, ret_stats, title, file_tag, strip=None, silent_plot=False):

        style = self._create_style(title, title)
        # keys = self._strategy_group_benchmark_ret_stats.keys()
        yoy = []

        for key in ret_stats.keys():
            col = ret_stats[key].yoy_rets()
            col.columns = [key]
            yoy.append(col)

        calculations = Calculations()
        ret_stats = calculations.pandas_outer_join(yoy)
        ret_stats.index = ret_stats.index.year

        ret_stats = self._strip_dataframe(ret_stats, strip)

        # ret_stats = ret_stats.sort_index()
        style.file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (' + file_tag + ') ' \
                            + str(style.scale_factor) + '.png'
        style.html_file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (' + file_tag + ') ' \
                                 + str(style.scale_factor) + '.html'
        style.display_brand_label = False
        style.date_formatter = "%Y"

        chart = Chart(ret_stats * 100, engine=self.DEFAULT_PLOT_ENGINE, chart_type='bar', style=style)
        if not (silent_plot): chart.plot()
        return chart

    def plot_strategy_group_leverage(self, silent_plot=False, reduce_plot=True, resample='B'):

        style = self._create_style("Leverage", "Group Leverage", reduce_plot=reduce_plot)

        chart = Chart(self._reduce_plot(self._strategy_group_leverage,
                                        reduce_plot=reduce_plot, resample=resample), engine=self.DEFAULT_PLOT_ENGINE,
                      chart_type='line', style=style)
        if not (silent_plot): chart.plot()
        return chart

    ###### plot signals and trades, in terms of units, notionals and contract sizes (eg. for futures)
    def plot_strategy_signals(self, date=None, strip=None, silent_plot=False):
        self._plot_signal(self._strategy_signal, label="positions (% portfolio notional)", caption="Positions",
                          date=date, strip=strip, silent_plot=silent_plot)

    def plot_strategy_trades(self, date=None, strip=None, silent_plot=False):
        self._plot_signal(self._strategy_trade, label="trades (% portfolio notional)", caption="Trades",
                          date=date, strip=strip, silent_plot=silent_plot)

    def plot_strategy_signals_notional(self, date=None, strip=None, silent_plot=False):
        self._plot_signal(self._strategy_signal_notional, label="positions (scaled by notional)", caption="Positions",
                          date=date, strip=strip, silent_plot=silent_plot)

    def plot_strategy_trades_notional(self, date=None, strip=None, silent_plot=False):
        self._plot_signal(self._strategy_trade_notional, label="trades (scaled by notional)", caption="Trades",
                          date=date, strip=strip, silent_plot=silent_plot)

    def plot_strategy_trades_notional_sizes(self, strip=None, silent_plot=False):

        if strip is not None: self._strategy_trade_notional_sizes.index = [k.replace(strip, '')
                                                                           for k in
                                                                           self._strategy_trade_notional_sizes.index]

        style = self._create_style("", "", reduce_plot=False)

        try:
            style.file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (Strategy trade notional size).png'
            style.html_file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (Strategy trade notional size).html'

            chart = Chart(self._strip_dataframe(self._strategy_trade_notional_sizes, strip),
                          engine=self.DEFAULT_PLOT_ENGINE,
                          chart_type='bar', style=style)
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_signals_contracts(self, date=None, strip=None, silent_plot=False):
        self._plot_signal(self._strategy_signal_contracts, label="positions (contracts)", caption="Positions",
                          date=date, strip=strip, silent_plot=silent_plot)

    def plot_strategy_trades_contracts(self, date=None, strip=None, silent_plot=False):
        self._plot_signal(self._strategy_trade_contracts, label="trades (contracts)", caption="Contracts",
                          date=date, strip=strip, silent_plot=silent_plot)

    ###### plot aggregated portfolio exposures
    def plot_strategy_total_exposures(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        df = pandas.concat([self._strategy_total_longs, self._strategy_total_shorts, self._strategy_total_exposure],
                           axis=1)

        try:
            chart = Chart(
                self._strip_dataframe(self._reduce_plot(df, reduce_plot=reduce_plot, resample=resample), strip),
                engine=self.DEFAULT_PLOT_ENGINE, chart_type='line',
                style=self._create_style("", "Strategy Total Exposures"))
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_net_exposures(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        try:
            chart = Chart(self._strip_dataframe(self._reduce_plot(self._strategy_net_exposure,
                                                                  reduce_plot=reduce_plot, resample=resample), strip),
                          engine=self.DEFAULT_PLOT_ENGINE, chart_type='line',
                          style=self._create_style("", "Strategy Net Exposures", reduce_plot=reduce_plot))
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_total_exposures_notional(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        df = pandas.concat([self._strategy_total_longs_notional,
                            self._strategy_total_shorts_notional, self._strategy_total_exposure_notional], axis=1)

        try:
            chart = Chart(
                self._strip_dataframe(self._reduce_plot(df / 1000000.0, reduce_plot=reduce_plot, resample=resample),
                                      strip),
                engine=self.DEFAULT_PLOT_ENGINE, chart_type='line',
                style=self._create_style("(mm)", "Strategy Total Exposures (mm)", reduce_plot=reduce_plot))
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    def plot_strategy_net_exposures_notional(self, strip=None, silent_plot=False, reduce_plot=True, resample='B'):

        try:
            chart = Chart(self._strip_dataframe(self._reduce_plot(self._strategy_net_exposure_notional / 1000000.0,
                                                                  reduce_plot=reduce_plot, resample=resample),
                                                strip),
                          engine=self.DEFAULT_PLOT_ENGINE, chart_type='line',
                          style=self._create_style("(mm)", "Strategy Net Exposures (mm)", reduce_plot=reduce_plot))
            if not (silent_plot): chart.plot()
            return chart
        except:
            pass

    #### grab signals for specific days
    def _grab_signals(self, strategy_signal, date=None, strip=None):
        if date is None:
            last_day = strategy_signal.ix[-1].transpose().to_frame()
        else:
            if not (isinstance(date, list)):
                date = [date]

            last_day = []

            for d in date:
                last_day.append(strategy_signal.ix[d].transpose().to_frame())

            last_day = pandas.concat(last_day, axis=1)
            last_day = last_day.sort_index(axis=1)

        if strip is not None:
            last_day.index = [x.replace(strip, '') for x in last_day.index]

        return last_day

    def _plot_signal(self, sig, label=' ', caption='', date=None, strip=None, silent_plot=False):

        ######## plot signals
        strategy_signal = 100 * (sig)
        last_day = self._grab_signals(strategy_signal, date=date, strip=strip)

        style = self._create_style(label, caption)

        chart = Chart(last_day, engine=self.DEFAULT_PLOT_ENGINE, chart_type='bar', style=style)
        if not (silent_plot): chart.plot()
        return chart

    def _strip_dataframe(self, data_frame, strip):
        if strip is None:
            return data_frame

        data_frame.columns = [x.replace(strip, '') for x in data_frame.columns]

        return data_frame

    def _create_style(self, title, file_add, reduce_plot=True):
        style = copy.deepcopy(self.CHART_STYLE)

        if self.SHOW_TITLES:
            style.title = self.FINAL_STRATEGY + " " + title
            
        style.display_legend = True
        style.scale_factor = self.SCALE_FACTOR
        style.source = self.CHART_SOURCE
        style.silent_display = not (self.SHOW_CHARTS)

        # when plotting many points use WebGl version of plotly if specified
        if not (reduce_plot):
            style.plotly_webgl = True

        if self.DEFAULT_PLOT_ENGINE not in ['plotly', 'cufflinks'] and self.SAVE_FIGURES:
            style.file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (' + file_add + ') ' \
                                + str(style.scale_factor) + '.png'

        style.html_file_output = self.DUMP_PATH + self.FINAL_STRATEGY + ' (' + file_add + ') ' \
                                 + str(style.scale_factor) + '.html'

        try:
            style.silent_display = self.SILENT_DISPLAY
        except:
            pass

        return style


#######################################################################################################################

class PortfolioModel(object):
    pass


#######################################################################################################################

class RiskEngine(object):
    """Adjusts signal weighting according to risk constraints (volatility targeting and position limit constraints)

    """

    def calculate_vol_adjusted_index_from_prices(self, prices_df, br):
        """Adjusts an index of prices for a vol target

        Parameters
        ----------
        br : BacktestRequest
            Parameters for the backtest specifying start date, finish data, transaction costs etc.
        asset_a_df : pandas.DataFrame
            Asset prices to be traded

        Returns
        -------
        pandas.Dataframe containing vol adjusted index
        """

        calculations = Calculations()

        returns_df, leverage_df = self.calculate_vol_adjusted_returns(prices_df, br, returns=False)

        return calculations.create_mult_index(returns_df)

    def calculate_vol_adjusted_returns(self, returns_df, br, returns=True):
        """Adjusts returns for a vol target

        Parameters
        ----------
        br : BacktestRequest
            Parameters for the backtest specifying start date, finish data, transaction costs etc.
        returns_a_df : pandas.DataFrame
            Asset returns to be traded

        Returns
        -------
        pandas.DataFrame
        """

        calculations = Calculations()

        if not returns: returns_df = calculations.calculate_returns(returns_df)

        leverage_df = self.calculate_leverage_factor(returns_df,
                                                     br.portfolio_vol_target, br.portfolio_vol_max_leverage,
                                                     br.portfolio_vol_periods, br.portfolio_vol_obs_in_year,
                                                     br.portfolio_vol_rebalance_freq, br.portfolio_vol_resample_freq,
                                                     br.portfolio_vol_resample_type,
                                                     period_shift=br.portfolio_vol_period_shift)

        vol_returns_df = calculations.calculate_signal_returns_with_tc_matrix(leverage_df, returns_df, tc=br.spot_tc_bp)
        vol_returns_df.columns = returns_df.columns

        return vol_returns_df, leverage_df

    def calculate_leverage_factor(self, returns_df, vol_target, vol_max_leverage, vol_periods=60, vol_obs_in_year=252,
                                  vol_rebalance_freq='BM', resample_freq=None, resample_type='mean',
                                  returns=True, period_shift=0):
        """Calculates the time series of leverage for a specified vol target

        Parameters
        ----------
        returns_df : DataFrame
            Asset returns
        vol_target : float
            vol target for assets
        vol_max_leverage : float
            maximum leverage allowed
        vol_periods : int
            number of periods to calculate volatility
        vol_obs_in_year : int
            number of observations in the year
        vol_rebalance_freq : str
            how often to rebalance
        resample_type : str
            do we need to resample the underlying data first? (eg. have we got intraday data?)
        returns : boolean
            is this returns time series or prices?
        period_shift : int
            should we delay the signal by a number of periods?

        Returns
        -------
        pandas.Dataframe
        """

        calculations = Calculations()
        filter = Filter()

        if resample_freq is not None:
            return
            # TODO not implemented yet

        if not returns: returns_df = calculations.calculate_returns(returns_df)

        roll_vol_df = calculations.rolling_volatility(returns_df,
                                                      periods=vol_periods, obs_in_year=vol_obs_in_year).shift(
            period_shift)

        # calculate the leverage as function of vol target (with max lev constraint)
        lev_df = vol_target / roll_vol_df

        if vol_max_leverage is not None:
            lev_df[lev_df > vol_max_leverage] = vol_max_leverage

        if resample_type is not None:
            lev_df = filter.resample_time_series_frequency(lev_df, vol_rebalance_freq, resample_type)

            returns_df, lev_df = calculations.join_left_fill_right(returns_df, lev_df)

        # # in case leverage changes on a weekend do outer join, and fill down
        # # the leverage
        # returns_df_1, lev_df = returns_df.align(lev_df, join='outer', axis=0)
        # lev_df = lev_df.fillna(method='ffill')
        #
        # # now realign back to days when we trade
        # returns_df, lev_df = returns_df.align(lev_df, join='left', axis=0)

        lev_df.ix[0:vol_periods] = numpy.nan  # ignore the first elements before the vol window kicks in

        return lev_df

    def calculate_position_clip_adjustment(self, portfolio_net_exposure, portfolio_total_exposure, br):
        """Calculate the leverage adjustment that needs to be made in the portfolio such that either the net exposure or
        the absolute exposure fits within predefined limits

        Parameters
        ----------
        portfolio_net_exposure : DataFrame
            Net exposure of the whole portfolio
        portfolio_total_exposure : DataFrame
            Absolute exposure of the whole portfolio
        br : BacktestRequest
            Includes parameters for setting position limits

        Returns
        -------
        DataFrame
        """

        position_clip_adjustment = None

        # adjust leverage of portfolio based on max NET position sizes
        if br.max_net_exposure is not None:
            portfolio_net_exposure = portfolio_net_exposure.shift(br.position_clip_period_shift)

            # add further constraints on portfolio (total net amount of longs and short)
            position_clip_adjustment = pandas.DataFrame(data=numpy.ones(len(portfolio_net_exposure.index)),
                                                        index=portfolio_net_exposure.index,
                                                        columns=['Portfolio'])

            portfolio_abs_exposure = portfolio_net_exposure.abs()

            # for those periods when the absolute net positioning is greater than our limit cut down the leverage
            position_clip_adjustment[(portfolio_abs_exposure > br.max_net_exposure).values] = \
                br.max_net_exposure / portfolio_abs_exposure

        # adjust leverage of portfolio based on max TOTAL position sizes
        if br.max_abs_exposure is not None:
            portfolio_abs_exposure = portfolio_abs_exposure.shift(br.position_clip_period_shift)

            # add further constraints on portfolio (total net amount of longs and short)
            position_clip_adjustment = pandas.DataFrame(data=numpy.ones(len(portfolio_abs_exposure.index)),
                                                        index=portfolio_abs_exposure.index,
                                                        columns=['Portfolio'])

            # for those periods when the absolute TOTAL positioning is greater than our limit cut down the leverage
            position_clip_adjustment[(portfolio_total_exposure > br.max_abs_exposure).values] = \
                br.max_abs_exposure / portfolio_total_exposure

        # only allow the position clip adjustment to change on certain days (eg. 'BM' = month end)
        if br.position_clip_rebalance_freq is not None:
            calculations = Calculations()
            filter = Filter()

            position_clip_adjustment = filter.resample_time_series_frequency(position_clip_adjustment,
                                                                             br.position_clip_rebalance_freq,
                                                                             br.position_clip_resample_type)

            a, position_clip_adjustment = calculations.join_left_fill_right(portfolio_net_exposure,
                                                                            position_clip_adjustment)

        return position_clip_adjustment



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

"""
MarketConstants

Has various constants required for the finmarketpy project. These have been defined as static variables.

"""



class MarketConstants(object):
    import platform

    ###### SETUP ENVIRONMENT VARIABLES ######
    plat = str(platform.platform()).lower()

    if 'linux' in plat: generic_plat = 'linux'
    elif 'windows' in plat: generic_plat = 'windows'

    # "thread" or "multiprocessing" (experimental!) library to use when backtesting
    backtest_thread_technique = "multiprocessing"

    multiprocessing_library = 'multiprocess'  # 'multiprocessing_on_dill' or 'multiprocess' or 'multiprocessing' or 'pathos'

    # how many threads to use for backtesting (don't do too many on slow machines!)
    # also some data sources will complain if you start too many parallel threads to call data!
    # for some data providers might get better performance from 1 thread only!
    backtest_thread_no = {'linux': 14,
                          'windows' : 1}

    hdf5_file_econ_file = "x"

    db_database_econ_file = ''
    db_server = '127.0.0.1'
    db_port = '27017'

    db_username = 'admin_root'
    db_password = 'TOFILL'

    write_engine = 'arctic'

    # overwrite field variables with those listed in MarketCred
    def __init__(self):
        try:
            from finmarketpy.util.marketcred import MarketCred
            cred_keys = MarketCred.__dict__.keys()

            for k in MarketConstants.__dict__.keys():
                if k in cred_keys and '__' not in k:
                    setattr(MarketConstants, k, getattr(MarketCred, k))
        except:
            pass


__author__ = 'saeedamen'

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import pandas

from findatapy.util.loggermanager import LoggerManager


class MarketLiquidity(object):
    """Calculates spread between bid/ask and also tick count.

    """

    def __init__(self):
        self.logger = LoggerManager().getLogger(__name__)
        return

    def calculate_spreads(self, data_frame, asset, bid_field = 'bid', ask_field = 'ask'):
        if isinstance(asset, str): asset = [asset]

        cols = [x + '.spread' for x in asset]

        data_frame_spreads = pandas.DataFrame(index=data_frame.index, columns=cols)

        for a in asset:
            data_frame_spreads[a + '.spread'] = data_frame[a + "." + ask_field] - data_frame[a + "." + bid_field]

        return data_frame_spreads

    def calculate_tick_count(self, data_frame, asset, freq = '1h'):
        if isinstance(asset, str): asset = [asset]

        data_frame_tick_count = data_frame.resample(freq, how='count').dropna()
        data_frame_tick_count = data_frame_tick_count[[0]]

        data_frame_tick_count.columns = [x + '.event' for x in asset]

        return data_frame_tick_count

if __name__ == '__main__':
    # see examples
    pass


__author__ = 'saeedamen'

# loading data
import datetime

from chartpy import Chart, Style
from findatapy.market import Market, MarketDataGenerator, MarketDataRequest

from findatapy.util.loggermanager import LoggerManager

logger = LoggerManager().getLogger(__name__)

chart = Chart(engine='matplotlib')

market = Market(market_data_generator=MarketDataGenerator())

# choose run_example = 0 for everything
# run_example = 1 - download BoE data from quandl

run_example = 0

###### fetch data from Quandl for BoE rate (using Bloomberg data)
if run_example == 1 or run_example == 0:
    # Monthly average of UK resident monetary financial institutions' (excl. Central Bank) sterling
    # Weighted average interest rate, other loans, new advances, on a fixed rate to private non-financial corporations (in percent)
    # not seasonally adjusted
    md_request = MarketDataRequest(
        start_date="01 Jan 2000",  # start date
        data_source='quandl',  # use Quandl as data source
        tickers=['Weighted interest rate'],
        fields=['close'],  # which fields to download
        vendor_tickers=['BOE/CFMBJ84'],  # ticker (Bloomberg)
        vendor_fields=['close'],  # which Bloomberg fields to download
        cache_algo='internet_load_return')  # how to return data

    df = market.fetch_market(md_request)

    style = Style()

    style.title = 'BoE weighted interest rate'
    style.scale_factor = 3
    style.file_output = "boe-rate.png"
    style.source = 'Quandl/BoE'

    chart.plot(df, style=style)



__author__ = 'saeedamen' # Saeed Amen / saeed@thalesians.com

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


import math

from pandas.tseries.offsets import CustomBusinessDay

from findatapy.timeseries import Calculations, Timezone, Filter, Calendar

class EventStudy(object):
    """Provides functions for doing event studies on price action on an intraday basis and daily basis.

    """

    def __init__(self):
        self.logger = LoggerManager().getLogger(__name__)
        return

    def get_economic_event_ret_over_custom_event_day(self, data_frame_in, event_dates, name, event, start, end, lagged = False,
                                              NYC_cutoff = 10):

        filter = Filter()
        event_dates = filter.filter_time_series_by_date(start, end, event_dates)

        data_frame = data_frame_in.copy(deep=True) # because we change the dates!

        timezone = Timezone()
        calendar = Calendar()

        bday = CustomBusinessDay(weekmask='Mon Tue Wed Thu Fri')

        event_dates_nyc = timezone.convert_index_from_UTC_to_new_york_time(event_dates)
        average_hour_nyc = numpy.average(event_dates_nyc.index.hour)

        event_dates = calendar.floor_date(event_dates)

        # realised is traditionally on later day eg. 3rd Jan realised ON is 2nd-3rd Jan realised
        # so if Fed meeting is on 2nd Jan later, then we need realised labelled on 3rd (so minus a day)
        # implied expires on next day eg. 3rd Jan implied ON is 3rd-4th Jan implied

        # TODO smarter way of adjusting dates, as sometimes events can be before/after 10am NY cut
        if (lagged and average_hour_nyc >= NYC_cutoff):
            data_frame.index = data_frame.index - bday
        elif (not lagged and average_hour_nyc < NYC_cutoff): # ie. implied
            data_frame.index = data_frame.index + bday

        # set as New York time and select only those ON vols at the 10am NY cut just before the event
        data_frame_events = data_frame.ix[event_dates.index]
        data_frame_events.columns = data_frame.columns.values + '-' + name + ' ' + event

        return data_frame_events

    def get_intraday_moves_over_custom_event(self, data_frame_rets, ef_time_frame, vol=False,
                                             minute_start = 5, mins = 3 * 60, min_offset = 0 , create_index = False,
                                             resample = False, freq = 'minutes', cumsum = True):

        filter = Filter()

        ef_time_frame = filter.filter_time_series_by_date(data_frame_rets.index[0], data_frame_rets.index[-1], ef_time_frame)
        ef_time = ef_time_frame.index

        if freq == 'minutes':
            ef_time_start = ef_time - timedelta(minutes = minute_start)
            ef_time_end = ef_time + timedelta(minutes = mins)
            ann_factor = 252 * 1440
        elif freq == 'days':
            ef_time = ef_time_frame.index.normalize()
            ef_time_start = ef_time - timedelta(days = minute_start)
            ef_time_end = ef_time + timedelta(days = mins)
            ann_factor = 252

        ords = range(-minute_start + min_offset, mins + min_offset)

        # all data needs to be equally spaced
        if resample:

            # make sure time series is properly sampled at 1 min intervals
            data_frame_rets = data_frame_rets.resample('1min')
            data_frame_rets = data_frame_rets.fillna(value = 0)
            data_frame_rets = filter.remove_out_FX_out_of_hours(data_frame_rets)

        data_frame_rets['Ind'] = numpy.nan

        start_index = data_frame_rets.index.searchsorted(ef_time_start)
        finish_index = data_frame_rets.index.searchsorted(ef_time_end)

        # not all observation windows will be same length (eg. last one?)

        # fill the indices which represent minutes
        # TODO vectorise this!
        for i in range(0, len(ef_time_frame.index)):
            try:
                data_frame_rets.ix[start_index[i]:finish_index[i], 'Ind'] = ords
            except:
                data_frame_rets.ix[start_index[i]:finish_index[i], 'Ind'] = ords[0:(finish_index[i] - start_index[i])]

        # set the release dates
        data_frame_rets.ix[start_index,'Rel'] = ef_time                                         # set entry points
        data_frame_rets.ix[finish_index + 1,'Rel'] = numpy.zeros(len(start_index))              # set exit points
        data_frame_rets['Rel'] = data_frame_rets['Rel'].fillna(method = 'pad')                  # fill down signals

        data_frame_rets = data_frame_rets[pandas.notnull(data_frame_rets['Ind'])]               # get rid of other

        data_frame = data_frame_rets.pivot(index='Ind',
                                           columns='Rel', values=data_frame_rets.columns[0])

        data_frame.index.names = [None]

        if create_index:
            calculations = Calculations()
            data_frame.ix[-minute_start + min_offset,:] = numpy.nan
            data_frame = calculations.create_mult_index(data_frame)
        else:
            if vol is True:
                # annualise (if vol)
                data_frame = data_frame.rolling(center=False,window=5).std() * math.sqrt(ann_factor)
            elif cumsum:
                data_frame = data_frame.cumsum()

        return data_frame

    def get_surprise_against_intraday_moves_over_custom_event(
            self, data_frame_cross_orig, ef_time_frame, cross, event_fx, event_name, start, end,
            offset_list = [1, 5, 30, 60], add_surprise = False, surprise_field = 'survey-average', freq = 'minutes'):

        ticker = event_fx + "-" + event_name + ".release-date-time-full"

        data_frame_agg = None
        data_frame_cross_orig = data_frame_cross_orig.resample('T').mean()  # resample to minute freq - in case there are missing values

        ef_time_start = ef_time_frame[ticker] - timedelta(minutes=1)        # start time
        indices_start = data_frame_cross_orig.index.isin(ef_time_start)

        for offset in offset_list:
            data_frame_cross = data_frame_cross_orig

            ef_time = ef_time_frame[ticker] + timedelta(minutes=offset - 1)     # end time

            # calculate returns over the x min event
            indices = data_frame_cross.index.isin(ef_time)
            col_dates = data_frame_cross.index[indices]

            col_rets = (data_frame_cross.iloc[indices].values) \
                       / (data_frame_cross.iloc[indices_start].values) - 1

            mkt_moves = pandas.DataFrame(index=col_dates)
            mkt_moves[cross + " " + str(offset) + "m move"] = col_rets
            mkt_moves.index.name = ticker
            mkt_moves.index = col_dates - timedelta(minutes=offset - 1)

            data_frame = ef_time_frame.join(mkt_moves, on=ticker, how="inner")
            temp_index = data_frame[ticker]

            spot_moves_list = []

            if data_frame_agg is None:
                data_frame_agg = data_frame
            else:
                label = cross + " " + str(offset) + "m move"
                spot_moves_list.append(label)
                data_frame = data_frame[label]
                data_frame.index = temp_index
                data_frame_agg = data_frame_agg.join(data_frame, on=ticker, how="inner")

        if add_surprise == True:
            data_frame_agg[event_fx + "-" + event_name + ".surprise"] = data_frame_agg[event_fx + "-" + event_name + ".actual-release"] \
                               - data_frame_agg[event_fx + "-" + event_name + "." + surprise_field]

        return data_frame_agg

########################################################################################################################

import datetime
from datetime import timedelta

import numpy

from finmarketpy.util.marketconstants import MarketConstants
from findatapy.market import IOEngine
from findatapy.util import ConfigManager
from findatapy.market import SpeedCache

try:
    from numbapro import autojit
except: pass

marketconstants = MarketConstants()

class EventsFactory(EventStudy):
    """Provides methods to fetch data on economic data events and to perform basic event studies for market data around
    these events. Note, requires a file of input of the following (transposed as columns!) - we give an example for
    NFP released on 7 Feb 2003 (note, that release-date-time-full, need not be fully aligned by row).

    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.Date	                31/01/2003 00:00
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.close	                xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.actual-release	        143
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.survey-median	        xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.survey-average	        xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.survey-high	        xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.survey-low	            xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.survey-high.1	        xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.number-observations	xyz
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.first-revision	        185
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.first-revision-date	20030307
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.release-dt	            20030207
    USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.release-date-time-full	08/01/1999 13:30

    """

    # _econ_data_frame = None

    # where your HDF5 file is stored with economic data
    # TODO integrate with on the fly downloading!
    _hdf5_file_econ_file = MarketConstants().hdf5_file_econ_file
    _db_database_econ_file = MarketConstants().db_database_econ_file

    ### manual offset for certain events where Bloomberg/data vendor displays the wrong date (usually because of time differences)
    _offset_events = {'AUD-Australia Labor Force Employment Change SA.release-dt' : 1}

    def __init__(self, df = None):
        super(EventStudy, self).__init__()

        self.config = ConfigManager()
        self.logger = LoggerManager().getLogger(__name__)
        self.filter = Filter()
        self.io_engine = IOEngine()
        self.speed_cache = SpeedCache()

        if df is not None:
            self._econ_data_frame = df
        else:
            self.load_economic_events()

        return

    def load_economic_events(self):
        self._econ_data_frame = self.speed_cache.get_dataframe(self._db_database_econ_file)

        if self._econ_data_frame is None:
            # self._econ_data_frame = self.io_engine.read_time_series_cache_from_disk(self._hdf5_file_econ_file)
            self._econ_data_frame = self.io_engine.read_time_series_cache_from_disk(
                self._db_database_econ_file, engine=marketconstants.write_engine,
                db_server=marketconstants.db_server,
                db_port=marketconstants.db_port,
                username=marketconstants.db_username,
                password=marketconstants.db_password)

            self.speed_cache.put_dataframe(self._db_database_econ_file, self._econ_data_frame)

    def harvest_category(self, category_name):
        cat = self.config.get_categories_from_tickers_selective_filter(category_name)

        for k in cat:
            md_request = self.market_data_generator.populate_md_request(k)
            data_frame = self.market_data_generator.fetch_market_data(md_request)

            # TODO allow merge of multiple sources

        return data_frame

    def get_economic_events(self):
        return self._econ_data_frame

    def dump_economic_events_csv(self, path):
        self._econ_data_frame.to_csv(path)

    def get_economic_event_date_time(self, name, event = None, csv = None):
        ticker = self.create_event_desciptor_field(name, event, "release-date-time-full")

        if csv is None:
            data_frame = self._econ_data_frame[ticker]
            data_frame.index = self._econ_data_frame[ticker]
        else:
            dateparse = lambda x: datetime.datetime.strptime(x, '%d/%m/%Y %H:%M')

            data_frame = pandas.read_csv(csv, index_col=0, parse_dates = True, date_parser=dateparse)

        data_frame = data_frame[pandas.notnull(data_frame.index)]

        start_date = datetime.datetime.strptime("01-Jan-1971", "%d-%b-%Y")
        self.filter.filter_time_series_by_date(start_date, None, data_frame)

        return data_frame

    def get_economic_event_date_time_dataframe(self, name, event = None, csv = None):
        series = self.get_economic_event_date_time(name, event, csv)

        data_frame = pandas.DataFrame(series.values, index=series.index)
        data_frame.columns.name = self.create_event_desciptor_field(name, event, "release-date-time-full")

        return data_frame

    def get_economic_event_date_time_fields(self, fields, name, event = None):
        ### acceptible fields
        # observation-date <- observation time for the index
        # actual-release
        # survey-median
        # survey-average
        # survey-high
        # survey-low
        # survey-high
        # number-observations
        # release-dt
        # release-date-time-full
        # first-revision
        # first-revision-date

        ticker = []

        # construct tickers of the form USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA.actual-release
        for i in range(0, len(fields)):
            ticker.append(self.create_event_desciptor_field(name, event, fields[i]))

        # index on the release-dt field eg. 20101230 (we shall convert this later)
        ticker_index = self.create_event_desciptor_field(name, event, "release-dt")

        ######## grab event date/times
        event_date_time = self.get_economic_event_date_time(name, event)
        date_time_fore = event_date_time.index

        # create dates for join later
        date_time_dt = [datetime.datetime(
                                date_time_fore[x].year,
                                date_time_fore[x].month,
                                date_time_fore[x].day)
                                for x in range(len(date_time_fore))]

        event_date_time_frame = pandas.DataFrame(event_date_time.index, date_time_dt)
        event_date_time_frame.index = date_time_dt

        ######## grab event date/fields
        self._econ_data_frame[name + ".observation-date"] = self._econ_data_frame.index
        data_frame = self._econ_data_frame[ticker]

        data_frame.index = self._econ_data_frame[ticker_index]

        data_frame = data_frame[data_frame.index != 0]              # eliminate any 0 dates (artifact of Excel)
        data_frame = data_frame[pandas.notnull(data_frame.index)]   # eliminate any NaN dates (artifact of Excel)
        ind_dt = data_frame.index

        # convert yyyymmdd format to datetime
        data_frame.index = [datetime.datetime(
                               int((ind_dt[x] - (ind_dt[x] % 10000))/10000),
                               int(((ind_dt[x] % 10000) - (ind_dt[x] % 100))/100),
                               int(ind_dt[x] % 100)) for x in range(len(ind_dt))]

        # HACK! certain events need an offset because BBG have invalid dates
        if ticker_index in self._offset_events:
             data_frame.index = data_frame.index + timedelta(days=self._offset_events[ticker_index])

        ######## join together event dates/date-time/fields in one data frame
        data_frame = event_date_time_frame.join(data_frame, how='inner')
        data_frame.index = pandas.to_datetime(data_frame.index)
        data_frame.index.name = ticker_index

        return data_frame

    def create_event_desciptor_field(self, name, event, field):
        if event is None:
            return name + "." + field
        else:
            return name + "-" + event + "." + field

    def get_all_economic_events_date_time(self):
        event_names = self.get_all_economic_events()
        columns = ['event-name', 'release-date-time-full']

        data_frame = pandas.DataFrame(data=numpy.zeros((0,len(columns))), columns=columns)

        for event in event_names:
            event_times = self.get_economic_event_date_time(event)

            for time in event_times:
                data_frame.append({'event-name':event, 'release-date-time-full':time}, ignore_index=True)

        return data_frame

    def get_all_economic_events(self):
        field_names = self._econ_data_frame.columns.values

        event_names = [x.split('.')[0] for x in field_names if '.Date' in x]

        event_names_filtered = [x for x in event_names if len(x) > 4]

        # sort list alphabetically (and remove any duplicates)
        return list(set(event_names_filtered))

    def get_economic_event_date(self, name, event = None):
        return self._econ_data_frame[
            self.create_event_desciptor_field(name, event, ".release-dt")]

    def get_economic_event_ret_over_custom_event_day(self, data_frame_in, name, event, start, end, lagged = False,
                                              NYC_cutoff = 10):

        # get the times of events
        event_dates = self.get_economic_event_date_time(name, event)

        return super(EventsFactory, self).get_economic_event_ret_over_custom_event_day(data_frame_in, event_dates, name, event, start, end,
                                                                                       lagged = lagged, NYC_cutoff = NYC_cutoff)

    def get_economic_event_vol_over_event_day(self, vol_in, name, event, start, end, realised = False):

        return self.get_economic_event_ret_over_custom_event_day(vol_in, name, event, start, end,
            lagged = realised)

        # return super(EventsFactory, self).get_economic_event_ret_over_event_day(vol_in, name, event, start, end, lagged = realised)

    def get_daily_moves_over_event(self):
        # TODO
        pass

    # return only US events etc. by dates
    def get_intraday_moves_over_event(self, data_frame_rets, cross, event_fx, event_name, start, end, vol, mins = 3 * 60,
                                      min_offset = 0, create_index = False, resample = False, freq = 'minutes'):

        ef_time_frame = self.get_economic_event_date_time_dataframe(event_fx, event_name)
        ef_time_frame = self.filter.filter_time_series_by_date(start, end, ef_time_frame)

        return self.get_intraday_moves_over_custom_event(data_frame_rets, ef_time_frame,
                                                         vol, mins = mins, min_offset = min_offset,
                                                         create_index = create_index, resample = resample, freq = freq)#, start, end)

    def get_surprise_against_intraday_moves_over_event(self, data_frame_cross_orig, cross, event_fx, event_name, start, end,
                                                       offset_list = [1, 5, 30, 60], add_surprise = False,
                                                       surprise_field = 'survey-average'):

        fields = ['actual-release', 'survey-median', 'survey-average', 'survey-high', 'survey-low']

        ef_time_frame = self.get_economic_event_date_time_fields(fields, event_fx, event_name)
        ef_time_frame = self.filter.filter_time_series_by_date(start, end, ef_time_frame)

        return self.get_surprise_against_intraday_moves_over_custom_event(data_frame_cross_orig, ef_time_frame, cross, event_fx, event_name, start, end,
                                                       offset_list = offset_list, add_surprise = add_surprise,
                                                       surprise_field = surprise_field)

########################################################################################################################

"""
HistEconDataFactory

Provides functions for getting historical economic data. Uses aliases for tickers, to make it relatively easy to use,
rather than having to remember all the underlying vendor tickers. Can use Fred, Quandl or Bloomberg.

The files below, contain default tickers and country groups. However, you can add whichever tickers you'd like.
- conf/all_econ_tickers.csv
- conf/econ_country_codes.csv
- conf/econ_country_groups.csv

These can be automatically generated via conf/econ_tickers.xlsm

"""

import pandas

from findatapy.market import MarketDataGenerator, MarketDataRequest
from findatapy.util import DataConstants, LoggerManager

class HistEconDataFactory(object):

    def __init__(self, market_data_generator = None):
        self.logger = LoggerManager().getLogger(__name__)

        self._all_econ_tickers = pandas.read_csv(DataConstants().all_econ_tickers)
        self._econ_country_codes = pandas.read_csv(DataConstants().econ_country_codes)
        self._econ_country_groups = pandas.read_csv(DataConstants().econ_country_groups)

        if market_data_generator is None:
            self.market_data_generator = MarketDataGenerator()
        else:
            self.market_data_generator = market_data_generator

    def get_economic_data_history(self, start_date, finish_date, country_group, data_type,
        source = 'fred', cache_algo = "internet_load_return"):

        #vendor_country_codes = self.fred_country_codes[country_group]
        #vendor_pretty_country = self.fred_nice_country_codes[country_group]

        if isinstance(country_group, list):
            pretty_country_names = country_group
        else:
            # get all the country names in the country_group
            pretty_country_names = list(self._econ_country_groups[
                self._econ_country_groups["Country Group"] == country_group]['Country'])

        # construct the pretty tickers
        pretty_tickers = [x + '-' + data_type for x in pretty_country_names]

        # get vendor tickers
        vendor_tickers = []

        for pretty_ticker in pretty_tickers:
            vendor_ticker = list(self._all_econ_tickers[
                                         self._all_econ_tickers["Full Code"] == pretty_ticker][source].values)

            if vendor_ticker == []:
                vendor_ticker = None
                self.logger.error('Could not find match for ' + pretty_ticker)
            else:
                vendor_ticker = vendor_ticker[0]

            vendor_tickers.append(vendor_ticker)

        vendor_fields = ['close']

        if source == 'bloomberg': vendor_fields = ['PX_LAST']

        md_request = MarketDataRequest(
                start_date = start_date,                            # start date
                finish_date = finish_date,                          # finish date
                category = 'economic',
                freq = 'daily',                                     # intraday data
                data_source = source,                               # use Bloomberg as data source
                cut = 'LOC',
                tickers = pretty_tickers,
                fields = ['close'],                                 # which fields to download
                vendor_tickers = vendor_tickers,
                vendor_fields = vendor_fields,                      # which Bloomberg fields to download
                cache_algo = cache_algo)                            # how to return data

        return self.market_data_generator.fetch_market_data(md_request)

    def grasp_coded_entry(self, df, index):
        df = df.ix[index:].stack()
        df = df.reset_index()
        df.columns = ['Date', 'Name', 'Val']

        countries = df['Name']

        countries = [x.split('-', 1)[0] for x in countries]

        df['Code'] = sum(
            [list(self._econ_country_codes[self._econ_country_codes["Country"] == x]['Code']) for x in countries],
            [])

        return df


from setuptools import setup, find_packages

long_description = """finmarketpy is a Python based library that enables you to analyze market data and also to backtest 
trading strategies using a simple to use API, which has prebuilt templates for you to define backtest."""

setup(name='finmarketpy',
      version='0.11.2',
      description='finmarketpy is a Python based library for backtesting trading strategies',
      author='Saeed Amen',
      author_email='saeed@cuemacro.com',
      license='Apache 2.0',
      long_description=long_description,
      keywords = ['trading', 'markets', 'currencies', 'pandas', 'data', 'Bloomberg', 'tick', 'stocks', 'equities'],
      url = 'https://github.com/cuemacro/finmarketpy',
      packages = find_packages(),
      include_package_data = True,
      install_requires = ['pandas',
                          'twython',
                          'pytz',
                          'requests',
                          'numpy',
                          'multiprocess',
                          'seasonal'],
	  zip_safe=False)


__author__ = 'saeedamen'

from finmarketpy.economics.eventstudy import EventsFactory, EventStudy, HistEconDataFactory
from finmarketpy.economics.marketliquidity import MarketLiquidity
from finmarketpy.economics.seasonality import Seasonality
from finmarketpy.economics.report import Report
from finmarketpy.economics.techindicator import TechIndicator
from finmarketpy.economics.techindicator import TechParams


__author__ = 'saeedamen'


# loading data
import datetime

from chartpy import Chart, Style
from finmarketpy.backtest import TradeAnalysis
from findatapy.market import Market, MarketDataGenerator, MarketDataRequest

from chartpy.style import Style
from findatapy.timeseries import Calculations
from findatapy.util.loggermanager import LoggerManager

ta = TradeAnalysis()
calc = Calculations()
logger = LoggerManager().getLogger(__name__)

chart = Chart(engine='matplotlib')

market = Market(market_data_generator=MarketDataGenerator())

# choose run_example = 0 for everything
# run_example = 1 - use PyFolio to analyse gold's return properties

run_example = 0

###### use PyFolio to analyse gold's return properties
if run_example == 1 or run_example == 0:
    md_request = MarketDataRequest(
                start_date = "01 Jan 1996",                         # start date
                data_source = 'bloomberg',                          # use Bloomberg as data source
                tickers = ['Gold'],
                fields = ['close'],                                 # which fields to download
                vendor_tickers = ['XAUUSD Curncy'],                 # ticker (Bloomberg)
                vendor_fields = ['PX_LAST'],                        # which Bloomberg fields to download
                cache_algo = 'internet_load_return')                # how to return data

    df = market.fetch_market(md_request)

    ta.run_strategy_returns_stats(None, index=df, engine='pyfolio')


__author__ = 'saeedamen & mhockenberger'  # Saeed Amen & Marcel Hockenberger

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import pandas as pd
import numpy as np
from collections import OrderedDict

from findatapy.timeseries import Calculations
from findatapy.util.loggermanager import LoggerManager


class TechIndicator(object):
    """Calculates various technical indicators and associated trading signals.

    Signals can also be filtered to allow only longs or short and also to flip signals.

    It is possible to create subclasses of
    TechIndicator to implement your own custom signals and technical indicators.

    At present it implements the following technical indicators (and associated signals)
        SMA - simple moving averages - buy if spot above SMA, sell if below)
        EMA - exponential moving average - buy if spot spot above EMA, sell if below)
        ROC - rate of change - buy if rate of change positive, sell if negative
        polarity - buy if input is positive, sell if negative
        SMA2 - double simple moving average - buy if faster SMA above slower SMA, sell if below
        RSI - relative strength indicator - buy if RSI exits overbought, sell if RSI exits oversold
        BB - Bollinger bands - buy if spot above upper Bollinger band, sell if below lower Bollinger Band
        ATR - Average True Range - useful for creating stop/take profit levels
        long_only - long only signal

    """

    def __init__(self):
        self.logger = LoggerManager().getLogger(__name__)
        self._techind = None
        self._signal = None

    def create_tech_ind(
            self,
            data_frame_non_nan,
            name,
            tech_params,
            data_frame_non_nan_early=None):
        self._signal = None
        self._techind = None

        if tech_params.fillna:
            data_frame = data_frame_non_nan.fillna(method="ffill")
        else:
            data_frame = data_frame_non_nan

        if data_frame_non_nan_early is not None:
            data_frame_early = data_frame_non_nan_early.fillna(method="ffill")

        if name == "SMA":

            if (data_frame_non_nan_early is not None):
                # calculate the lagged sum of the n-1 point
                if pd.__version__ < '0.17':
                    rolling_sum = pd.rolling_sum(
                        data_frame.shift(1).rolling,
                        window=tech_params.sma_period - 1)
                else:
                    rolling_sum = data_frame.shift(1).rolling(
                        center=False, window=tech_params.sma_period - 1).sum()

                # add non-nan one for today
                rolling_sum = rolling_sum + data_frame_early

                # calculate average = sum / n
                self._techind = rolling_sum / tech_params.sma_period

                narray = np.where(data_frame_early > self._techind, 1, -1)
            else:
                if pd.__version__ < '0.17':
                    self._techind = pd.rolling_sum(
                        data_frame, window=tech_params.sma_period)
                else:
                    self._techind = data_frame.rolling(
                        window=tech_params.sma_period, center=False).mean()

                narray = np.where(data_frame > self._techind, 1, -1)

            self._signal = pd.DataFrame(index=data_frame.index, data=narray)
            self._signal.loc[0:tech_params.sma_period] = np.nan
            self._signal.columns = [
                x + " SMA Signal" for x in data_frame.columns.values]

            self._techind.columns = [
                x + " SMA" for x in data_frame.columns.values]

        elif name == "EMA":

            # self._techind = pd.ewma(data_frame, span = tech_params.ema_period)
            self._techind = data_frame.ewm(
                ignore_na=False,
                span=tech_params.ema_period,
                min_periods=0,
                adjust=True).mean()

            narray = np.where(data_frame > self._techind, 1, -1)

            self._signal = pd.DataFrame(index=data_frame.index, data=narray)
            self._signal.loc[0:tech_params.ema_period] = np.nan
            self._signal.columns = [
                x + " EMA Signal" for x in data_frame.columns.values]

            self._techind.columns = [
                x + " EMA" for x in data_frame.columns.values]

        elif name == "ROC":

            if (data_frame_non_nan_early is not None):
                self._techind = data_frame_early / \
                    data_frame.shift(tech_params.roc_period) - 1
            else:
                self._techind = data_frame / \
                    data_frame.shift(tech_params.roc_period) - 1

            narray = np.where(self._techind > 0, 1, -1)

            self._signal = pd.DataFrame(index=data_frame.index, data=narray)
            self._signal.loc[0:tech_params.roc_period] = np.nan
            self._signal.columns = [
                x + " ROC Signal" for x in data_frame.columns.values]

            self._techind.columns = [
                x + " ROC" for x in data_frame.columns.values]

        elif name == "polarity":
            self._techind = data_frame

            narray = np.where(self._techind > 0, 1, -1)

            self._signal = pd.DataFrame(index=data_frame.index, data=narray)
            self._signal.columns = [
                x + " Polarity Signal" for x in data_frame.columns.values]

            self._techind.columns = [
                x + " Polarity" for x in data_frame.columns.values]

        elif name == "SMA2":
            sma = data_frame.rolling(
                window=tech_params.sma_period,
                center=False).mean()
            sma2 = data_frame.rolling(
                window=tech_params.sma2_period,
                center=False).mean()

            narray = np.where(sma > sma2, 1, -1)

            self._signal = pd.DataFrame(index=data_frame.index, data=narray)
            self._signal.columns = [
                x + " SMA2 Signal" for x in data_frame.columns.values]

            sma.columns = [x + " SMA" for x in data_frame.columns.values]
            sma2.columns = [x + " SMA2" for x in data_frame.columns.values]
            most = max(tech_params.sma_period, tech_params.sma2_period)
            self._signal.loc[0:most] = np.nan
            self._techind = pd.concat([sma, sma2], axis=1)

        elif name in ['RSI']:
            # delta = data_frame.diff()
            #
            # dUp, dDown = delta.copy(), delta.copy()
            # dUp[dUp < 0] = 0
            # dDown[dDown > 0] = 0
            #
            # rolUp = pd.rolling_mean(dUp, tech_params.rsi_period)
            # rolDown = pd.rolling_mean(dDown, tech_params.rsi_period).abs()
            #
            # rsi = rolUp / rolDown

            # Get the difference in price from previous step
            delta = data_frame.diff()
            # Get rid of the first row, which is NaN since it did not have a previous
            # row to calculate the differences
            delta = delta[1:]

            # Make the positive gains (up) and negative gains (down) Series
            up, down = delta.copy(), delta.copy()
            up[up < 0] = 0
            down[down > 0] = 0

            # Calculate the EWMA
            roll_up1 = pd.stats.moments.ewma(up, tech_params.rsi_period)
            roll_down1 = pd.stats.moments.ewma(
                down.abs(), tech_params.rsi_period)

            # Calculate the RSI based on EWMA
            RS1 = roll_up1 / roll_down1
            RSI1 = 100.0 - (100.0 / (1.0 + RS1))

            # Calculate the SMA
            roll_up2 = up.rolling(
                window=tech_params.rsi_period,
                center=False).mean()
            roll_down2 = down.abs().rolling(
                window=tech_params.rsi_period, center=False).mean()

            # Calculate the RSI based on SMA
            RS2 = roll_up2 / roll_down2
            RSI2 = 100.0 - (100.0 / (1.0 + RS2))

            self._techind = RSI2
            self._techind.columns = [
                x + " RSI" for x in data_frame.columns.values]

            signal = data_frame.copy()

            sells = (signal.shift(-1) <
                     tech_params.rsi_lower) & (signal > tech_params.rsi_lower)
            buys = (signal.shift(-1) >
                    tech_params.rsi_upper) & (signal < tech_params.rsi_upper)

            # print (buys[buys == True])

            # buys
            signal[buys] = 1
            signal[sells] = -1
            signal[~(buys | sells)] = np.nan
            signal = signal.fillna(method='ffill')

            self._signal = signal

            self._signal.loc[0:tech_params.rsi_period] = np.nan
            self._signal.columns = [
                x + " RSI Signal" for x in data_frame.columns.values]

        elif name in ["BB"]:
            # calcuate Bollinger bands
            mid = data_frame.rolling(
                center=False, window=tech_params.bb_period).mean()
            mid.columns = [x + " BB Mid" for x in data_frame.columns.values]
            std_dev = data_frame.rolling(
                center=False, window=tech_params.bb_period).std()
            BB_std = tech_params.bb_mult * std_dev

            lower = pd.DataFrame(
                data=mid.values - BB_std.values,
                index=mid.index,
                columns=data_frame.columns)

            upper = pd.DataFrame(
                data=mid.values + BB_std.values,
                index=mid.index,
                columns=data_frame.columns)

            # calculate signals
            signal = data_frame.copy()

            buys = signal > upper
            sells = signal < lower

            signal[buys] = 1
            signal[sells] = -1
            signal[~(buys | sells)] = np.nan
            signal = signal.fillna(method='ffill')

            self._signal = signal
            self._signal.loc[0:tech_params.bb_period] = np.nan
            self._signal.columns = [
                x + " " + name + " Signal" for x in data_frame.columns.values]

            lower.columns = [
                x + " BB Lower" for x in data_frame.columns.values]
            upper.columns = [x + " BB Mid" for x in data_frame.columns.values]
            upper.columns = [
                x + " BB Lower" for x in data_frame.columns.values]

            self._techind = pd.concat([lower, mid, upper], axis=1)
        elif name == "long-only":
            # have +1 signals only
            self._techind = data_frame  # the technical indicator is just "prices"

            narray = np.ones((len(data_frame.index), len(data_frame.columns)))

            self._signal = pd.DataFrame(index=data_frame.index, data=narray)
            self._signal.columns = [
                x + " Long Only Signal" for x in data_frame.columns.values]

            self._techind.columns = [
                x + " Long Only" for x in data_frame.columns.values]

        elif name == "ATR":
            # get all the asset names (assume we have names 'close', 'low', 'high' in the Data)
            # keep ordering of assets
            asset_name = list(OrderedDict.fromkeys(
                [x.split('.')[0] for x in data_frame.columns]))

            df = []

            # can improve the performance of this if vectorise more!
            for a in asset_name:

                close = [a + '.close']
                low = [a + '.low']
                high = [a + '.high']

                # if we don't fill NaNs, we need to remove those rows and then
                # calculate the ATR
                if not(tech_params.fillna):
                    data_frame_short = data_frame[[close[0], low[0], high[0]]]
                    data_frame_short = data_frame_short.dropna()
                else:
                    data_frame_short = data_frame

                prev_close = data_frame_short[close].shift(1)

                c1 = data_frame_short[high].values - \
                    data_frame_short[low].values
                c2 = np.abs(data_frame_short[high].values - prev_close.values)
                c3 = np.abs(data_frame_short[low].values - prev_close.values)

                true_range = np.max((c1, c2, c3), axis=0)
                true_range = pd.DataFrame(
                    index=data_frame_short.index, data=true_range, columns=[
                        close[0] + ' True Range'])

                # put back NaNs into ATR if necessary
                if (not(tech_params.fillna)):
                    true_range = true_range.reindex(
                        data_frame.index, fill_value=np.nan)

                df.append(true_range)

            calc = Calculations()
            true_range = calc.pandas_outer_join(df)

            self._techind = true_range.rolling(
                window=tech_params.atr_period, center=False).mean()
            # self._techind = true_range.ewm(ignore_na=False, span=tech_params.atr_period, min_periods=0, adjust=True).mean()

            self._techind.columns = [x + ".close ATR" for x in asset_name]

        elif name in ["VWAP"]:
            asset_name = list(OrderedDict.fromkeys(
                [x.split('.')[0] for x in data_frame.columns]))

            df = []

            for a in asset_name:
                high = [a + '.high']
                low = [a + '.low']
                close = [a + '.close']
                volume = [a + '.volume']

                if not tech_params.fillna:
                    df_mod = data_frame[[high[0], low[0], close[0], volume[0]]]
                    df_mod.dropna(inplace=True)
                else:
                    df_mod = data_frame

                l = df_mod[low].values
                h = df_mod[high].values
                c = df_mod[close].values
                v = df_mod[volume].values

                vwap = np.cumsum(((h + l + c) / 3) * v) / np.cumsum(v)
                vwap = pd.DataFrame(index=df_mod.index, data=vwap,
                                    columns=[close[0] + ' VWAP'])
                print(vwap.columns)

                if not tech_params.fillna:
                    vwap = vwap.reindex(data_frame.index, fill_value=np.nan)

                df.append(vwap)

            calc = Calculations()
            vwap = calc.pandas_outer_join(df)

            self._techind = vwap

            self._techind.columns = [x + ".close VWAP" for x in asset_name]

        self.create_custom_tech_ind(
            data_frame_non_nan,
            name,
            tech_params,
            data_frame_non_nan_early)

        # TODO create other indicators
        if hasattr(tech_params, 'only_allow_longs'):
            self._signal[self._signal < 0] = 0

        # TODO create other indicators
        if hasattr(tech_params, 'only_allow_shorts'):
            self._signal[self._signal > 0] = 0

        # apply signal multiplier (typically to flip signals)
        if hasattr(tech_params, 'signal_mult'):
            self._signal = self._signal * tech_params.signal_mult

        if hasattr(tech_params, 'strip_signal_name'):
            if tech_params.strip_signal_name:
                self._signal.columns = data_frame.columns

        return self._techind, self._signal

    def create_custom_tech_ind(
            self,
            data_frame_non_nan,
            name,
            tech_params,
            data_frame_non_nan_early):
        return

    def get_techind(self):
        return self._techind

    def get_signal(self):
        return self._signal

##########################################################################


class TechParams:
    """Holds parameters for calculation of technical indicators.

    """

    def __init__(self, fillna=True, atr_period=14, sma_period=20,
                 green_n=4, green_count=9, red_n=2, red_count=13):
        self.fillna = fillna
        self.atr_period = atr_period
        self.sma_period = sma_period

        self.green_n = green_n
        self.green_count = green_count
        self.red_n = red_n
        self.red_count = red_count

    @property
    def fillna(self): return self.__fillna

    @fillna.setter
    def fillna(self, fillna): self.__fillna = fillna

    @property
    def atr_period(self): return self.__atr_period

    @atr_period.setter
    def atr_period(self, atr_period): self.__atr_period = atr_period

    @property
    def sma_period(self): return self.__sma_period

    @sma_period.setter
    def sma_period(self, sma_period): self.__sma_period = sma_period

    @property
    def green_n(self): return self.__green_n

    @green_n.setter
    def green_n(self, green_n): self.__green_n = green_n

    @property
    def green_count(self): return self.__green_count

    @green_count.setter
    def green_count(self, green_count): self.__green_count = green_count

    @property
    def red_n(self): return self.__red_n

    @red_n.setter
    def red_n(self, red_n): self.__red_n = red_n

    @property
    def red_count(self): return self.__red_count

    @red_count.setter
    def red_count(self, red_count): self.__red_count = red_count

    # TODO add specific fields so can error check fields



__author__ = 'saeedamen'


# loading data
import datetime

from chartpy import Chart, Style
from findatapy.market import Market, MarketDataGenerator, MarketDataRequest
from finmarketpy.economics import TechIndicator, TechParams

from findatapy.util.loggermanager import LoggerManager

logger = LoggerManager().getLogger(__name__)

chart = Chart(engine='matplotlib')

market = Market(market_data_generator=MarketDataGenerator())
tech_ind = TechIndicator()

# choose run_example = 0 for everything
# run_example = 1 - download S&P500 from Quandl, calculate ATR and plot

run_example = 0

###### fetch data from Quandl for BoE rate (using Bloomberg data)
if run_example == 1 or run_example == 0:

    # downloaded S&P500
    md_request = MarketDataRequest(
                start_date = "01 Jan 2000",                         # start date
                data_source = 'quandl',                             # use Quandl as data source
                tickers = ['S&P500'],
                fields = ['close', 'open', 'high', 'low'],          # which fields to download
                vendor_tickers = ['YAHOO/INDEX_GSPC'],              # ticker (Bloomberg)
                vendor_fields = ['close', 'open', 'high', 'low'],   # which Bloomberg fields to download
                cache_algo = 'internet_load_return')                # how to return data

    df = market.fetch_market(md_request)

    print(df)

    tech_params = TechParams()
    tech_params.atr_period = 14
    tech_ind.create_tech_ind(df, 'ATR', tech_params)

    style = Style()

    style.title = 'S&P500 ATR'
    style.scale_factor = 2
    style.file_output = "sp500.png"
    style.source = 'Quandl/Yahoo'

    df = tech_ind.get_techind()

    chart.plot(df, style=style)